# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/data/user_deployment_settings.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from tecton_proto.common import secret_pb2 as tecton__proto_dot_common_dot_secret__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='tecton_proto/data/user_deployment_settings.proto',
  package='tecton_proto.data',
  syntax='proto2',
  serialized_options=b'\n\017com.tecton.dataP\001',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n0tecton_proto/data/user_deployment_settings.proto\x12\x11tecton_proto.data\x1a tecton_proto/common/secret.proto\"\xa3\x02\n\x16UserDeploymentSettings\x12G\n user_deployment_settings_version\x18\x01 \x01(\x05R\x1duserDeploymentSettingsVersion\x12R\n\x11\x64\x61tabricks_config\x18\x02 \x01(\x0b\x32#.tecton_proto.data.DatabricksConfigH\x00R\x10\x64\x61tabricksConfig\x12T\n\x13user_spark_settings\x18\x03 \x01(\x0b\x32$.tecton_proto.data.UserSparkSettingsR\x11userSparkSettingsB\x16\n\x14\x64\x61ta_platform_config\"q\n\x10\x44\x61tabricksConfig\x12#\n\rworkspace_url\x18\x01 \x01(\tR\x0cworkspaceUrl\x12\x38\n\tapi_token\x18\x02 \x01(\x0b\x32\x1b.tecton_proto.common.SecretR\x08\x61piToken\"\xd7\x01\n\x11UserSparkSettings\x12\x30\n\x14instance_profile_arn\x18\x01 \x01(\tR\x12instanceProfileArn\x12R\n\nspark_conf\x18\x02 \x03(\x0b\x32\x33.tecton_proto.data.UserSparkSettings.SparkConfEntryR\tsparkConf\x1a<\n\x0eSparkConfEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01\x42\x13\n\x0f\x63om.tecton.dataP\x01'
  ,
  dependencies=[tecton__proto_dot_common_dot_secret__pb2.DESCRIPTOR,])




_USERDEPLOYMENTSETTINGS = _descriptor.Descriptor(
  name='UserDeploymentSettings',
  full_name='tecton_proto.data.UserDeploymentSettings',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='user_deployment_settings_version', full_name='tecton_proto.data.UserDeploymentSettings.user_deployment_settings_version', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='userDeploymentSettingsVersion', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='databricks_config', full_name='tecton_proto.data.UserDeploymentSettings.databricks_config', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='databricksConfig', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='user_spark_settings', full_name='tecton_proto.data.UserDeploymentSettings.user_spark_settings', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='userSparkSettings', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='data_platform_config', full_name='tecton_proto.data.UserDeploymentSettings.data_platform_config',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=106,
  serialized_end=397,
)


_DATABRICKSCONFIG = _descriptor.Descriptor(
  name='DatabricksConfig',
  full_name='tecton_proto.data.DatabricksConfig',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='workspace_url', full_name='tecton_proto.data.DatabricksConfig.workspace_url', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='workspaceUrl', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='api_token', full_name='tecton_proto.data.DatabricksConfig.api_token', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='apiToken', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=399,
  serialized_end=512,
)


_USERSPARKSETTINGS_SPARKCONFENTRY = _descriptor.Descriptor(
  name='SparkConfEntry',
  full_name='tecton_proto.data.UserSparkSettings.SparkConfEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='tecton_proto.data.UserSparkSettings.SparkConfEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='key', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='tecton_proto.data.UserSparkSettings.SparkConfEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='value', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=670,
  serialized_end=730,
)

_USERSPARKSETTINGS = _descriptor.Descriptor(
  name='UserSparkSettings',
  full_name='tecton_proto.data.UserSparkSettings',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='instance_profile_arn', full_name='tecton_proto.data.UserSparkSettings.instance_profile_arn', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='instanceProfileArn', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='spark_conf', full_name='tecton_proto.data.UserSparkSettings.spark_conf', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='sparkConf', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_USERSPARKSETTINGS_SPARKCONFENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=515,
  serialized_end=730,
)

_USERDEPLOYMENTSETTINGS.fields_by_name['databricks_config'].message_type = _DATABRICKSCONFIG
_USERDEPLOYMENTSETTINGS.fields_by_name['user_spark_settings'].message_type = _USERSPARKSETTINGS
_USERDEPLOYMENTSETTINGS.oneofs_by_name['data_platform_config'].fields.append(
  _USERDEPLOYMENTSETTINGS.fields_by_name['databricks_config'])
_USERDEPLOYMENTSETTINGS.fields_by_name['databricks_config'].containing_oneof = _USERDEPLOYMENTSETTINGS.oneofs_by_name['data_platform_config']
_DATABRICKSCONFIG.fields_by_name['api_token'].message_type = tecton__proto_dot_common_dot_secret__pb2._SECRET
_USERSPARKSETTINGS_SPARKCONFENTRY.containing_type = _USERSPARKSETTINGS
_USERSPARKSETTINGS.fields_by_name['spark_conf'].message_type = _USERSPARKSETTINGS_SPARKCONFENTRY
DESCRIPTOR.message_types_by_name['UserDeploymentSettings'] = _USERDEPLOYMENTSETTINGS
DESCRIPTOR.message_types_by_name['DatabricksConfig'] = _DATABRICKSCONFIG
DESCRIPTOR.message_types_by_name['UserSparkSettings'] = _USERSPARKSETTINGS
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

UserDeploymentSettings = _reflection.GeneratedProtocolMessageType('UserDeploymentSettings', (_message.Message,), {
  'DESCRIPTOR' : _USERDEPLOYMENTSETTINGS,
  '__module__' : 'tecton_proto.data.user_deployment_settings_pb2'
  # @@protoc_insertion_point(class_scope:tecton_proto.data.UserDeploymentSettings)
  })
_sym_db.RegisterMessage(UserDeploymentSettings)

DatabricksConfig = _reflection.GeneratedProtocolMessageType('DatabricksConfig', (_message.Message,), {
  'DESCRIPTOR' : _DATABRICKSCONFIG,
  '__module__' : 'tecton_proto.data.user_deployment_settings_pb2'
  # @@protoc_insertion_point(class_scope:tecton_proto.data.DatabricksConfig)
  })
_sym_db.RegisterMessage(DatabricksConfig)

UserSparkSettings = _reflection.GeneratedProtocolMessageType('UserSparkSettings', (_message.Message,), {

  'SparkConfEntry' : _reflection.GeneratedProtocolMessageType('SparkConfEntry', (_message.Message,), {
    'DESCRIPTOR' : _USERSPARKSETTINGS_SPARKCONFENTRY,
    '__module__' : 'tecton_proto.data.user_deployment_settings_pb2'
    # @@protoc_insertion_point(class_scope:tecton_proto.data.UserSparkSettings.SparkConfEntry)
    })
  ,
  'DESCRIPTOR' : _USERSPARKSETTINGS,
  '__module__' : 'tecton_proto.data.user_deployment_settings_pb2'
  # @@protoc_insertion_point(class_scope:tecton_proto.data.UserSparkSettings)
  })
_sym_db.RegisterMessage(UserSparkSettings)
_sym_db.RegisterMessage(UserSparkSettings.SparkConfEntry)


DESCRIPTOR._options = None
_USERSPARKSETTINGS_SPARKCONFENTRY._options = None
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/args/basic_info.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from tecton_proto.args import diff_options_pb2 as tecton__proto_dot_args_dot_diff__options__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='tecton_proto/args/basic_info.proto',
  package='tecton_proto.args',
  syntax='proto2',
  serialized_options=b'\n\017com.tecton.argsP\001',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\"tecton_proto/args/basic_info.proto\x12\x11tecton_proto.args\x1a$tecton_proto/args/diff_options.proto\"\x86\x02\n\tBasicInfo\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12\'\n\x0b\x64\x65scription\x18\x02 \x01(\tB\x05\x92M\x02\x08\x01R\x0b\x64\x65scription\x12\x1b\n\x05owner\x18\x06 \x01(\tB\x05\x92M\x02\x08\x01R\x05owner\x12\x1d\n\x06\x66\x61mily\x18\x04 \x01(\tB\x05\x92M\x02\x08\x01R\x06\x66\x61mily\x12\x41\n\x04tags\x18\x05 \x03(\x0b\x32&.tecton_proto.args.BasicInfo.TagsEntryB\x05\x92M\x02\x08\x01R\x04tags\x1a\x37\n\tTagsEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01J\x04\x08\x03\x10\x04\x42\x13\n\x0f\x63om.tecton.argsP\x01'
  ,
  dependencies=[tecton__proto_dot_args_dot_diff__options__pb2.DESCRIPTOR,])




_BASICINFO_TAGSENTRY = _descriptor.Descriptor(
  name='TagsEntry',
  full_name='tecton_proto.args.BasicInfo.TagsEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='tecton_proto.args.BasicInfo.TagsEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='key', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='tecton_proto.args.BasicInfo.TagsEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='value', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=297,
  serialized_end=352,
)

_BASICINFO = _descriptor.Descriptor(
  name='BasicInfo',
  full_name='tecton_proto.args.BasicInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='tecton_proto.args.BasicInfo.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='name', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='description', full_name='tecton_proto.args.BasicInfo.description', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\222M\002\010\001', json_name='description', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='owner', full_name='tecton_proto.args.BasicInfo.owner', index=2,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\222M\002\010\001', json_name='owner', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='family', full_name='tecton_proto.args.BasicInfo.family', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\222M\002\010\001', json_name='family', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='tags', full_name='tecton_proto.args.BasicInfo.tags', index=4,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\222M\002\010\001', json_name='tags', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_BASICINFO_TAGSENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=96,
  serialized_end=358,
)

_BASICINFO_TAGSENTRY.containing_type = _BASICINFO
_BASICINFO.fields_by_name['tags'].message_type = _BASICINFO_TAGSENTRY
DESCRIPTOR.message_types_by_name['BasicInfo'] = _BASICINFO
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

BasicInfo = _reflection.GeneratedProtocolMessageType('BasicInfo', (_message.Message,), {

  'TagsEntry' : _reflection.GeneratedProtocolMessageType('TagsEntry', (_message.Message,), {
    'DESCRIPTOR' : _BASICINFO_TAGSENTRY,
    '__module__' : 'tecton_proto.args.basic_info_pb2'
    # @@protoc_insertion_point(class_scope:tecton_proto.args.BasicInfo.TagsEntry)
    })
  ,
  'DESCRIPTOR' : _BASICINFO,
  '__module__' : 'tecton_proto.args.basic_info_pb2'
  # @@protoc_insertion_point(class_scope:tecton_proto.args.BasicInfo)
  })
_sym_db.RegisterMessage(BasicInfo)
_sym_db.RegisterMessage(BasicInfo.TagsEntry)


DESCRIPTOR._options = None
_BASICINFO_TAGSENTRY._options = None
_BASICINFO.fields_by_name['description']._options = None
_BASICINFO.fields_by_name['owner']._options = None
_BASICINFO.fields_by_name['family']._options = None
_BASICINFO.fields_by_name['tags']._options = None
# @@protoc_insertion_point(module_scope)



from typing import Dict
from typing import Optional

from tecton.data_sources.base_data_source_config import BaseStreamDSConfig
from tecton_proto.args import data_source_pb2
from tecton_proto.args import virtual_data_source_pb2
from tecton_spark import function_serialization
from tecton_spark.time_utils import strict_pytimeparse


class KafkaDSConfig(BaseStreamDSConfig):
    """
    Configuration used to reference a Kafka stream.

    The KafkaDSConfig class is used to create a reference to a Kafka stream.

    This class used as an input to a :class:`StreamDataSource`'s parameter ``stream_ds_config``. This class is not
    a Tecton Object: it is a grouping of parameters. Declaring this class alone will not register a data source.
    Instead, declare as part of ``StreamDataSource`` that takes this configuration class instance as a parameter.

    """

    def __init__(
        self,
        kafka_bootstrap_servers: str,
        topics: str,
        raw_stream_translator,
        timestamp_key: str,
        default_watermark_delay_threshold: str = None,
        options: Optional[Dict[str, str]] = None,
        ssl_keystore_location: Optional[str] = None,
        ssl_keystore_password_secret_id: Optional[str] = None,
        ssl_truststore_location: Optional[str] = None,
        ssl_truststore_password_secret_id: Optional[str] = None,
        security_protocol: Optional[str] = None,
    ):
        """
        Instantiates a new KafkaDSConfig.

        :param kafka_bootstrap_servers: A comma-separated list of the Kafka bootstrap server addresses. Passed directly
                                        to the Spark ``kafka.bootstrap.servers`` option.
        :param topics: A comma-separated list of Kafka topics to subscribe to. Passed directly to the Spark ``subscribe``
                       option.
        :param raw_stream_translator: Python user defined function f(DataFrame) -> DataFrame that takes in raw
                                      Pyspark data source DataFrame and translates it to the DataFrame to be
                                      consumed by the Feature View. See an example of
                                      raw_stream_translator in the `User Guide`_.
        :param timestamp_key: Name of the column containing timestamp for watermarking.
        :param default_watermark_delay_threshold: (Optional) Watermark time interval, e.g: "24 hours", used by Spark Structured Streaming to account for late-arriving data. See: https://docs.tecton.ai/v2/overviews/framework/feature_views/stream_feature_view.html#productionizing-a-stream
        :param options: (Optional) A map of additional Spark readStream options
        :param ssl_keystore_location: An DBFS (Databricks only) or S3 URI that points to the keystore file that should be used for SSL brokers. Note for S3 URIs, this must be configured by your Tecton representative.
            Example: ``s3://tecton-${cluster_name}/kafka-credentials/kafka_client_keystore.jks``
            Example: ``dbfs:/kafka-credentials/kafka_client_keystore.jks``
        :param ssl_keystore_password_secret_id: The config key for the password for the Keystore.
            Should start with ``SECRET_``, example: ``SECRET_KAFKA_PRODUCTION``.
        :param ssl_truststore_location: An DBFS (Databricks only) or S3 URI that points to the truststore file that should be used for SSL brokers. Note for S3 URIs, this must be configured by your Tecton representative. If not provided, the default truststore for your compute provider will be used. Note that this is required for AWS-signed keystores on Databricks.
            Example: ``s3://tecton-${cluster_name}/kafka-credentials/kafka_client_truststore.jks``
            Example: ``dbfs:/kafka-credentials/kafka_client_truststore.jks``
        :param ssl_truststore_password_secret_id (Optional): The config key for the password for the Truststore.
            Should start with ``SECRET_``, example: ``SECRET_KAFKA_PRODUCTION``.
        :param security_protocol: (Optional) Security protocol passed to kafka.security.protocol. See Kafka documentation for valid values.

        :return: A KafkaDSConfig class instance.

        .. _User Guide: https://docs.tecton.ai/v2/overviews/framework/data_sources.html

        Example of a KafkaDSConfig declaration:

        .. code-block:: python

            import pyspark
            from tecton import KafkaDSConfig


            def raw_data_deserialization(df:pyspark.sql.DataFrame) -> pyspark.sql.DataFrame:
                from pyspark.sql.functions import from_json, col
                from pyspark.sql.types import StringType, TimestampType

                PAYLOAD_SCHEMA = (
                  StructType()
                        .add("accountId", StringType(), False)
                        .add("transaction_id", StringType(), False)
                )

                EVENT_ENVELOPE_SCHEMA = (
                  StructType()
                        .add("timestamp", TimestampType(), False)
                        .add("payload", PAYLOAD_SCHEMA, False)
                )

                value = col("value").cast("string")
                df = df.withColumn("event", from_json(value, EVENT_ENVELOPE_SCHEMA))
                df = df.withColumn("accountId", col("event.payload.accountId"))
                df = df.withColumn("transaction_id", col("event.payload.transaction_id"))
                df = df.withColumn("timestamp", col("event.timestamp"))

                return df

            # Declare Kafka DSConfig instance object that can be used as an argument in StreamDataSource
            click_stream_kafka_ds = KafkaDSConfig(default_watermark_delay_threshold="7 days",
                                        kafka_bootstrap_servers="127.0.0.1:12345",
                                        topics="click-events-json",
                                        timestamp_key="click_event_timestamp",
                                        raw_stream_translator=raw_data_deserialization)
        """
        self._args = args = data_source_pb2.KafkaDataSourceArgs()
        args.kafka_bootstrap_servers = kafka_bootstrap_servers
        args.topics = topics
        args.raw_stream_translator.CopyFrom(function_serialization.to_proto(raw_stream_translator))
        args.timestamp_key = timestamp_key
        if default_watermark_delay_threshold:
            args.default_watermark_delay_threshold.FromSeconds(strict_pytimeparse(default_watermark_delay_threshold))
        for key in sorted((options or {}).keys()):
            option = data_source_pb2.Option()
            option.key = key
            option.value = options[key]
            args.options.append(option)
        if ssl_keystore_location:
            args.ssl_keystore_location = ssl_keystore_location
        if ssl_keystore_password_secret_id:
            args.ssl_keystore_password_secret_id = ssl_keystore_password_secret_id
        if ssl_truststore_location:
            args.ssl_truststore_location = ssl_truststore_location
        if ssl_truststore_password_secret_id:
            args.ssl_truststore_password_secret_id = ssl_truststore_password_secret_id
        if security_protocol:
            args.security_protocol = security_protocol

    def _merge_stream_args(self, data_source_args: virtual_data_source_pb2.VirtualDataSourceArgs):
        data_source_args.kafka_ds_config.CopyFrom(self._args)



import inspect
import re
from pathlib import Path

from tecton import conf
from tecton_proto.args.repo_metadata_pb2 import SourceInfo

# Matches frame strings such as "<string>"
SKIP_FRAME_REGEX = re.compile("\<.*\>")


def get_current_workspace():
    return conf.get_or_none("TECTON_WORKSPACE")


def get_fco_source_info() -> SourceInfo:
    from tecton_spark.repo_file_handler import _maybe_get_repo_root

    source_info = SourceInfo()
    if not _maybe_get_repo_root():
        pass
    else:
        frames = inspect.stack()
        repo_root_path = Path(_maybe_get_repo_root())
        for frame in frames:
            if SKIP_FRAME_REGEX.match(frame.frame.f_code.co_filename) is not None:
                continue
            frame_path = Path(frame.frame.f_code.co_filename).resolve()
            if frame_path.exists() and (repo_root_path in frame_path.parents):
                rel_filename = frame_path.relative_to(repo_root_path)
                source_info.source_lineno = str(frame.lineno)
                source_info.source_filename = str(rel_filename)
                break
    return source_info


def get_debug(ctx):
    while ctx.parent:
        ctx = ctx.parent
    return ctx.params["debug"]


def get_verbose(ctx):
    while ctx.parent:
        ctx = ctx.parent
    return ctx.params["verbose"]



from typing import Mapping
from typing import Optional
from typing import Sequence
from typing import Union

import numpy as np
import pandas as pd
from pyspark.sql import DataFrame as pysparkDF

from tecton._internals import errors
from tecton._internals.feature_retrieval_internal import find_dependent_feature_set_items
from tecton._internals.utils import infer_timestamp
from tecton._internals.utils import is_bfc_mode_single
from tecton.interactive.data_frame import DataFrame
from tecton.interactive.data_frame import FeatureVector
from tecton.interactive.feature_service import FeatureService
from tecton.interactive.feature_view import FeatureView
from tecton.tecton_errors import TectonValidationError
from tecton_spark.feature_definition_wrapper import FeatureDefinitionWrapper as FeatureDefinition
from tecton_spark.feature_set_config import FeatureSetConfig
from tecton_spark.logger import get_logger

logger = get_logger("Tecton")


def get_online_features(
    features: Union[FeatureService, Sequence[FeatureView]],
    join_keys: Optional[Mapping[str, Union[int, np.int_, str, bytes]]] = None,
    include_join_keys_in_response: bool = False,
    request_data: Optional[Mapping[str, Union[int, np.int_, str, bytes, float]]] = None,
) -> FeatureVector:
    logger.warning(
        "Deprecated. Use 'FeatureService.get_online_features' or 'FeatureView.get_online_features' directly on the Feature Service or Feature View objects instead to fetch features from the Online Store. "
        + "See the api reference: https://docs.tecton.ai/api-reference/stubs/tecton.interactive.FeatureService.html#tecton.interactive.FeatureService.get_online_features"
    )
    if isinstance(features, FeatureService):
        return features.get_feature_vector(
            join_keys=join_keys,
            include_join_keys_in_response=include_join_keys_in_response,
            request_context_map=request_data,
        )
    elif isinstance(features, list):
        features_vector = FeatureVector([], [], [])
        for fv in features:
            features_vector._update(
                fv.get_feature_vector(
                    join_keys=join_keys,
                    include_join_keys_in_response=include_join_keys_in_response,
                    request_context_map=request_data,
                )
            )
        return features_vector
    else:
        raise TectonValidationError(f"Unexpected data type for features: {type(features)}")


def get_historical_features(
    features: Union[Sequence[FeatureView], FeatureService],
    spine: Union[pysparkDF, pd.DataFrame],
    from_source: bool = False,
    save: bool = False,
    save_as: str = None,
    timestamp_key: Optional[str] = None,
) -> DataFrame:
    logger.warning(
        "Deprecated. Use 'FeatureService.get_historical_features' or 'FeatureView.get_historical_features' directly on the Feature Service or Feature View objects instead to view historical values. "
        + "See the api reference: https://docs.tecton.ai/api-reference/stubs/tecton.interactive.FeatureService.html#tecton.interactive.FeatureService.get_historical_features"
    )
    if not timestamp_key:
        timestamp_key = infer_timestamp(spine)

    if isinstance(features, FeatureService):
        return features.get_historical_features(
            spine=spine, timestamp_key=timestamp_key, from_source=from_source, save=save, save_as=save_as
        )
    elif isinstance(features, list):
        for feature in features:
            if not isinstance(feature, FeatureView):
                raise TectonValidationError(
                    "The `features` parameter can only be a FeatureService, or a List of FeatureViews"
                )

        feature_set_config = FeatureSetConfig()
        for feature in features:
            fd = FeatureDefinition(feature._proto, feature._fco_container)

            if from_source and is_bfc_mode_single(fd):
                raise errors.FV_BFC_SINGLE_FROM_SOURCE

            if feature._proto.HasField("on_demand_feature_view"):
                inputs = find_dependent_feature_set_items(
                    feature._fco_container,
                    feature._proto.pipeline.root,
                    visited_inputs={},
                    fv_id=feature.id,
                    workspace_name=feature.workspace,
                )
                feature_set_config._definitions_and_configs = feature_set_config._definitions_and_configs + inputs
            feature_set_config._add(fd)
        return FeatureService._get_feature_dataframe_internal(
            feature_set_config, spine, timestamp_key, use_materialized_data=not from_source
        )
    else:
        raise TectonValidationError(
            "The `features` parameter can only be " "a FeatureService, or a List of FeatureViews"
        )



from typing import Optional

from tecton.data_sources.base_data_source_config import BaseBatchDSConfig
from tecton_proto.args import data_source_pb2
from tecton_proto.args import virtual_data_source_pb2
from tecton_spark import function_serialization


class SnowflakeDSConfig(BaseBatchDSConfig):
    """
    Configuration used to reference a Snowflake table or query.

    The SnowflakeDSConfig class is used to create a reference to a Snowflake table. You can also create a
    reference to a query on one or more tables, which will be registered in Tecton in a similar way as a view
    is registered in other data systems.

    This class used as an input to a :class:`BatchDataSource`'s parameter ``batch_ds_config``. This class is not
    a Tecton Object: it is a grouping of parameters. Declaring this class alone will not register a data source.
    Instead, declare as part of ``BatchDataSource`` that takes this configuration class instance as a parameter.
    """

    def __init__(
        self,
        *,
        database: str,
        schema: str,
        warehouse: Optional[str] = None,
        url: Optional[str] = None,
        role: Optional[str] = None,
        table: Optional[str] = None,
        query: Optional[str] = None,
        timestamp_key: Optional[str] = None,
        raw_batch_translator=None,
    ):
        """
        Instantiates a new SnowflakeDSConfig. One of table and query should be specified when creating this file.

        :param url: The connection URL to Snowflake, which contains account information
                         (e.g. https://xy12345.eu-west-1.snowflakecomputing.com).
        :param database: The Snowflake database for this Data source.
        :param schema: The Snowflake schema for this Data source.
        :param warehouse: The Snowflake warehouse for this Data source.
        :param role: (Optional) The Snowflake role that should be used for this Data source.

        :param table: The table for this Data source. Only one of `table` and `query` must be specified.
        :param query: The query for this Data source. Only one of `table` and `query` must be specified.

        :param raw_batch_translator: Python user defined function f(DataFrame) -> DataFrame that takes in raw
                                     PySpark data source DataFrame and translates it to the DataFrame to be
                                     consumed by the Feature View. See an example of
                                     raw_batch_translator in the `User Guide`_.
        :param timestamp_key: (Optional) The name of the timestamp column (after the raw_batch_translator has been applied).
                               The column name does not need to be specified if there is exactly one timestamp column after the translator is applied.
                               This is needed for efficient time filtering when materializing batch features.

        :return: A SnowflakeDSConfig class instance.

        .. _User Guide: https://docs.tecton.ai/v2/overviews/framework/data_sources.html

        Example of a SnowflakeDSConfig declaration:

        .. code-block:: python

            from tecton import SnowflakeDSConfig, BatchDataSource

            # Declare SnowflakeDSConfig instance object that can be used as an argument in BatchDataSource
            snowflake_ds_config = SnowflakeDSConfig(
                                              url="https://<your-cluster>.eu-west-1.snowflakecomputing.com/",
                                              database="CLICK_STREAM_DB",
                                              schema="CLICK_STREAM_SCHEMA",
                                              warehouse="COMPUTE_WH",
                                              table="CLICK_STREAM_FEATURES",
                                              query="SELECT timestamp as ts, created, user_id, clicks, click_rate"
                                                     "FROM CLICK_STREAM_DB.CLICK_STREAM_FEATURES")

            # Use in the BatchDataSource
            snowflake_ds = BatchDataSource(name="click_stream_snowflake_ds",
                                           batch_ds_config=snowflake_ds_config)
        """
        self._args = args = data_source_pb2.SnowflakeDataSourceArgs()
        args.database = database
        args.schema = schema
        if url:
            args.url = url
        if warehouse:
            args.warehouse = warehouse

        if role:
            args.role = role

        if table:
            args.table = table
        if query:
            args.query = query

        if raw_batch_translator is not None:
            args.raw_batch_translator.CopyFrom(function_serialization.to_proto(raw_batch_translator))
        if timestamp_key:
            args.timestamp_key = timestamp_key

    def _merge_batch_args(self, data_source_args: virtual_data_source_pb2.VirtualDataSourceArgs):
        data_source_args.snowflake_ds_config.CopyFrom(self._args)



from typing import List

from pyspark.sql import DataFrame
from pyspark.sql import functions

from tecton_proto.common import aggregation_function_pb2 as afpb
from tecton_proto.data.feature_view_pb2 import TrailingTimeWindowAggregation
from tecton_spark.aggregation_plans import get_aggregation_plan
from tecton_spark.aggregation_utils import get_aggregation_column_prefix_from_column_name
from tecton_spark.aggregation_utils import get_aggregation_column_prefixes
from tecton_spark.aggregation_utils import get_aggregation_function_name
from tecton_spark.spark_time_utils import convert_timestamp_to_epoch

TEMPORAL_ANCHOR_COLUMN_NAME = "_anchor_time"
WINDOW_COLUMN_NAME = "window"


def _get_feature_partial_aggregations(aggregation_plan, feature_name: str):
    column_names = set()
    for column_name, aggregated_column in zip(
        aggregation_plan.materialized_column_names(feature_name),
        aggregation_plan.partial_aggregation_transform(feature_name),
    ):
        if column_name in column_names:
            continue
        column_names.add(column_name)

        yield column_name, aggregated_column.alias(column_name)


def _get_continuous_aggregation_value(column_prefix, feature_input_name):
    if column_prefix == get_aggregation_function_name(afpb.AGGREGATION_FUNCTION_COUNT):
        return functions.lit(1).cast("long")
    elif column_prefix == get_aggregation_function_name(afpb.AGGREGATION_FUNCTION_LASTN):
        return functions.array(functions.col(feature_input_name))
    elif column_prefix == get_aggregation_function_name(afpb.AGGREGATION_FUNCTION_MEAN):
        return functions.col(feature_input_name).cast("double")
    else:
        return functions.col(feature_input_name)


def _convert_window_to_anchor_time(
    output_df,
    is_continuous,
    time_key,
    version,
    window_start_column_name,
    window_end_column_name,
    convert_window_times_to_epoch,
):
    def _add_time_column(df, input_ts_column_name, output_column_name):
        col = functions.col(input_ts_column_name)
        return df.withColumn(
            output_column_name, convert_timestamp_to_epoch(col, version) if convert_window_times_to_epoch else col
        )

    # For continuous aggregations this will simply be the time key.
    if is_continuous:
        return _add_time_column(output_df, time_key, TEMPORAL_ANCHOR_COLUMN_NAME)

    # Grouping by Spark Window introduces the "window" struct with "start" and "end" columns.
    # We only need to keep the "start" column as an anchor time.
    anchor_column_name = window_start_column_name if window_start_column_name else TEMPORAL_ANCHOR_COLUMN_NAME
    output_df = _add_time_column(output_df, f"{WINDOW_COLUMN_NAME}.start", anchor_column_name)

    if window_end_column_name:
        output_df = _add_time_column(output_df, f"{WINDOW_COLUMN_NAME}.end", window_end_column_name)

    return output_df.drop(WINDOW_COLUMN_NAME)


def construct_partial_time_aggregation_df(
    df,
    join_keys: List[str],
    time_aggregation: TrailingTimeWindowAggregation,
    version,
    window_start_column_name: str = None,
    window_end_column_name: str = None,
    convert_to_epoch: bool = True,
):
    output_columns = set()
    if not time_aggregation.is_continuous:
        group_by_cols = [functions.col(join_key) for join_key in join_keys]
        slide_str = f"{time_aggregation.aggregation_slide_period.seconds} seconds"
        window_spec = functions.window(time_aggregation.time_key, slide_str, slide_str)
        group_by_cols = [window_spec] + group_by_cols
        aggregations = []
        for feature in time_aggregation.features:
            aggregation_plan = get_aggregation_plan(
                feature.function, feature.function_params, time_aggregation.is_continuous
            )
            for name, aggregation in _get_feature_partial_aggregations(aggregation_plan, feature.input_feature_name):
                if name in output_columns:
                    continue
                output_columns.add(name)
                aggregations.append(aggregation)
        output_df = df.groupBy(*group_by_cols).agg(*aggregations)
        # There isn't an Scala Encoder that works with a list directly, so instead we wrap the list in an object. Here
        # we strip the object to get just the list.
        for col_name in output_columns:
            if col_name.startswith("lastn"):
                output_df = output_df.withColumn(col_name, output_df[col_name].values)
    else:
        columns_to_drop = set()
        for feature in time_aggregation.features:
            column_prefixes = get_aggregation_column_prefixes(feature.function)
            for column_prefix in column_prefixes:
                full_name = f"{column_prefix}_{feature.input_feature_name}"
                if full_name in output_columns:
                    continue
                output_columns.add(full_name)
                df = df.withColumn(
                    full_name, _get_continuous_aggregation_value(column_prefix, feature.input_feature_name)
                )
            columns_to_drop.add(feature.input_feature_name)
        # Drop the original feature columns.
        for column in columns_to_drop:
            df = df.drop(column)
        output_df = df

    output_df = _convert_window_to_anchor_time(
        output_df,
        time_aggregation.is_continuous,
        time_aggregation.time_key,
        version,
        window_start_column_name,
        window_end_column_name,
        convert_to_epoch,
    )
    return output_df


def rename_partial_aggregate_columns(
    df: DataFrame, slide_interval_string: str, trailing_time_window_aggregation: TrailingTimeWindowAggregation
) -> DataFrame:
    """Rename partial aggregate columns to human readable format."""
    # Create a map from intermediate rollup column name to preferred column names.
    renaming_map = dict()
    for feature in trailing_time_window_aggregation.features:
        aggregation_plan = get_aggregation_plan(
            feature.function, feature.function_params, trailing_time_window_aggregation.is_continuous
        )
        for old_name in aggregation_plan.materialized_column_names(feature.input_feature_name):
            aggregation_function_name = get_aggregation_column_prefix_from_column_name(feature.function, old_name)
            renaming_map[old_name] = f"{feature.input_feature_name}_{aggregation_function_name}_{slide_interval_string}"
    for (old_name, new_name) in renaming_map.items():
        df = df.withColumnRenamed(old_name, new_name)
    return df



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/args/feature_service.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from tecton_proto.common import id_pb2 as tecton__proto_dot_common_dot_id__pb2
from tecton_proto.args import basic_info_pb2 as tecton__proto_dot_args_dot_basic__info__pb2
from tecton_proto.args import diff_options_pb2 as tecton__proto_dot_args_dot_diff__options__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='tecton_proto/args/feature_service.proto',
  package='tecton_proto.args',
  syntax='proto2',
  serialized_options=b'\n\017com.tecton.argsP\001',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\'tecton_proto/args/feature_service.proto\x12\x11tecton_proto.args\x1a\x1ctecton_proto/common/id.proto\x1a\"tecton_proto/args/basic_info.proto\x1a$tecton_proto/args/diff_options.proto\"d\n\x11LoggingConfigArgs\x12\x1f\n\x0bsample_rate\x18\x01 \x01(\x02R\nsampleRate\x12.\n\x13log_effective_times\x18\x02 \x01(\x08R\x11logEffectiveTimes\"V\n\nColumnPair\x12!\n\x0cspine_column\x18\x01 \x01(\tR\x0bspineColumn\x12%\n\x0e\x66\x65\x61ture_column\x18\x02 \x01(\tR\rfeatureColumn\"\xec\x01\n\x1c\x46\x65\x61tureServiceFeaturePackage\x12\x45\n\x12\x66\x65\x61ture_package_id\x18\x01 \x01(\x0b\x32\x17.tecton_proto.common.IdR\x10\x66\x65\x61turePackageId\x12K\n\x12override_join_keys\x18\x02 \x03(\x0b\x32\x1d.tecton_proto.args.ColumnPairR\x10overrideJoinKeys\x12\x1c\n\tnamespace\x18\x03 \x01(\tR\tnamespace\x12\x1a\n\x08\x66\x65\x61tures\x18\x04 \x03(\tR\x08\x66\x65\x61tures\"\xed\x02\n\x12\x46\x65\x61tureServiceArgs\x12\x45\n\x12\x66\x65\x61ture_service_id\x18\x01 \x01(\x0b\x32\x17.tecton_proto.common.IdR\x10\x66\x65\x61tureServiceId\x12\x37\n\x04info\x18\x02 \x01(\x0b\x32\x1c.tecton_proto.args.BasicInfoB\x05\x92M\x02\x10\x01R\x04info\x12Z\n\x10\x66\x65\x61ture_packages\x18\x03 \x03(\x0b\x32/.tecton_proto.args.FeatureServiceFeaturePackageR\x0f\x66\x65\x61turePackages\x12;\n\x16online_serving_enabled\x18\x04 \x01(\x08\x42\x05\x92M\x02\x08\x01R\x14onlineServingEnabled\x12>\n\x07logging\x18\x05 \x01(\x0b\x32$.tecton_proto.args.LoggingConfigArgsR\x07loggingB\x13\n\x0f\x63om.tecton.argsP\x01'
  ,
  dependencies=[tecton__proto_dot_common_dot_id__pb2.DESCRIPTOR,tecton__proto_dot_args_dot_basic__info__pb2.DESCRIPTOR,tecton__proto_dot_args_dot_diff__options__pb2.DESCRIPTOR,])




_LOGGINGCONFIGARGS = _descriptor.Descriptor(
  name='LoggingConfigArgs',
  full_name='tecton_proto.args.LoggingConfigArgs',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='sample_rate', full_name='tecton_proto.args.LoggingConfigArgs.sample_rate', index=0,
      number=1, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='sampleRate', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='log_effective_times', full_name='tecton_proto.args.LoggingConfigArgs.log_effective_times', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='logEffectiveTimes', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=166,
  serialized_end=266,
)


_COLUMNPAIR = _descriptor.Descriptor(
  name='ColumnPair',
  full_name='tecton_proto.args.ColumnPair',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='spine_column', full_name='tecton_proto.args.ColumnPair.spine_column', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='spineColumn', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='feature_column', full_name='tecton_proto.args.ColumnPair.feature_column', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='featureColumn', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=268,
  serialized_end=354,
)


_FEATURESERVICEFEATUREPACKAGE = _descriptor.Descriptor(
  name='FeatureServiceFeaturePackage',
  full_name='tecton_proto.args.FeatureServiceFeaturePackage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='feature_package_id', full_name='tecton_proto.args.FeatureServiceFeaturePackage.feature_package_id', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='featurePackageId', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='override_join_keys', full_name='tecton_proto.args.FeatureServiceFeaturePackage.override_join_keys', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='overrideJoinKeys', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='namespace', full_name='tecton_proto.args.FeatureServiceFeaturePackage.namespace', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='namespace', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='features', full_name='tecton_proto.args.FeatureServiceFeaturePackage.features', index=3,
      number=4, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='features', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=357,
  serialized_end=593,
)


_FEATURESERVICEARGS = _descriptor.Descriptor(
  name='FeatureServiceArgs',
  full_name='tecton_proto.args.FeatureServiceArgs',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='feature_service_id', full_name='tecton_proto.args.FeatureServiceArgs.feature_service_id', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='featureServiceId', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='info', full_name='tecton_proto.args.FeatureServiceArgs.info', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\222M\002\020\001', json_name='info', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='feature_packages', full_name='tecton_proto.args.FeatureServiceArgs.feature_packages', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='featurePackages', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='online_serving_enabled', full_name='tecton_proto.args.FeatureServiceArgs.online_serving_enabled', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\222M\002\010\001', json_name='onlineServingEnabled', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='logging', full_name='tecton_proto.args.FeatureServiceArgs.logging', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='logging', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=596,
  serialized_end=961,
)

_FEATURESERVICEFEATUREPACKAGE.fields_by_name['feature_package_id'].message_type = tecton__proto_dot_common_dot_id__pb2._ID
_FEATURESERVICEFEATUREPACKAGE.fields_by_name['override_join_keys'].message_type = _COLUMNPAIR
_FEATURESERVICEARGS.fields_by_name['feature_service_id'].message_type = tecton__proto_dot_common_dot_id__pb2._ID
_FEATURESERVICEARGS.fields_by_name['info'].message_type = tecton__proto_dot_args_dot_basic__info__pb2._BASICINFO
_FEATURESERVICEARGS.fields_by_name['feature_packages'].message_type = _FEATURESERVICEFEATUREPACKAGE
_FEATURESERVICEARGS.fields_by_name['logging'].message_type = _LOGGINGCONFIGARGS
DESCRIPTOR.message_types_by_name['LoggingConfigArgs'] = _LOGGINGCONFIGARGS
DESCRIPTOR.message_types_by_name['ColumnPair'] = _COLUMNPAIR
DESCRIPTOR.message_types_by_name['FeatureServiceFeaturePackage'] = _FEATURESERVICEFEATUREPACKAGE
DESCRIPTOR.message_types_by_name['FeatureServiceArgs'] = _FEATURESERVICEARGS
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

LoggingConfigArgs = _reflection.GeneratedProtocolMessageType('LoggingConfigArgs', (_message.Message,), {
  'DESCRIPTOR' : _LOGGINGCONFIGARGS,
  '__module__' : 'tecton_proto.args.feature_service_pb2'
  # @@protoc_insertion_point(class_scope:tecton_proto.args.LoggingConfigArgs)
  })
_sym_db.RegisterMessage(LoggingConfigArgs)

ColumnPair = _reflection.GeneratedProtocolMessageType('ColumnPair', (_message.Message,), {
  'DESCRIPTOR' : _COLUMNPAIR,
  '__module__' : 'tecton_proto.args.feature_service_pb2'
  # @@protoc_insertion_point(class_scope:tecton_proto.args.ColumnPair)
  })
_sym_db.RegisterMessage(ColumnPair)

FeatureServiceFeaturePackage = _reflection.GeneratedProtocolMessageType('FeatureServiceFeaturePackage', (_message.Message,), {
  'DESCRIPTOR' : _FEATURESERVICEFEATUREPACKAGE,
  '__module__' : 'tecton_proto.args.feature_service_pb2'
  # @@protoc_insertion_point(class_scope:tecton_proto.args.FeatureServiceFeaturePackage)
  })
_sym_db.RegisterMessage(FeatureServiceFeaturePackage)

FeatureServiceArgs = _reflection.GeneratedProtocolMessageType('FeatureServiceArgs', (_message.Message,), {
  'DESCRIPTOR' : _FEATURESERVICEARGS,
  '__module__' : 'tecton_proto.args.feature_service_pb2'
  # @@protoc_insertion_point(class_scope:tecton_proto.args.FeatureServiceArgs)
  })
_sym_db.RegisterMessage(FeatureServiceArgs)


DESCRIPTOR._options = None
_FEATURESERVICEARGS.fields_by_name['info']._options = None
_FEATURESERVICEARGS.fields_by_name['online_serving_enabled']._options = None
# @@protoc_insertion_point(module_scope)



import io
import os
import sys
import tarfile
import time
from pathlib import Path
from typing import Iterable
from typing import List
from typing import Mapping
from typing import Optional
from typing import Tuple
from typing import Union

import requests
from colorama import Fore
from colorama import Style
from google.protobuf.json_format import MessageToJson
from yaspin.spinners import Spinners

import tecton
from .cli_utils import confirm_or_exit
from .cli_utils import human_fco_type
from .error_utils import format_server_errors
from tecton import Entity
from tecton import FeatureService
from tecton._internals import errors
from tecton._internals import metadata_service
from tecton._internals.analytics import TectonStateUpdateMetrics
from tecton._internals.utils import is_live_workspace
from tecton.cli import printer
from tecton.cli.common import get_current_workspace
from tecton.data_sources.data_source import BaseDataSource
from tecton.feature_views.feature_view import FeatureDefinition
from tecton.transformations.builtin_transformations import BuiltinTransformation
from tecton.transformations.transformation import Transformation
from tecton_proto.args.entity_pb2 import EntityArgs
from tecton_proto.args.fco_args_pb2 import FcoArgs
from tecton_proto.args.feature_service_pb2 import FeatureServiceArgs
from tecton_proto.args.feature_view_pb2 import FeatureViewArgs
from tecton_proto.args.feature_view_pb2 import FeatureViewType
from tecton_proto.args.new_transformation_pb2 import NewTransformationArgs
from tecton_proto.args.repo_metadata_pb2 import FeatureRepoSourceInfo
from tecton_proto.args.virtual_data_source_pb2 import DataSourceType
from tecton_proto.args.virtual_data_source_pb2 import VirtualDataSourceArgs
from tecton_proto.cli import repo_diff_pb2
from tecton_proto.common.id_pb2 import Id
from tecton_proto.data import state_update_pb2
from tecton_proto.metadataservice.metadata_service_pb2 import ApplyStateUpdateRequest
from tecton_proto.metadataservice.metadata_service_pb2 import NewStateUpdateRequest
from tecton_proto.metadataservice.metadata_service_pb2 import QueryStateUpdateRequest
from tecton_spark.id_helper import IdHelper

UnwrappedFcoArgs = Union[
    VirtualDataSourceArgs,
    EntityArgs,
    NewTransformationArgs,
    FeatureViewArgs,
    FeatureServiceArgs,
]


_FCO_TYPE_TO_OBJECT_TYPE = {
    "virtual_data_source": repo_diff_pb2.TectonObjectType.DATA_SOURCE,
    "entity": repo_diff_pb2.TectonObjectType.ENTITY,
    "feature_view": repo_diff_pb2.TectonObjectType.FEATURE_VIEW,
    "feature_service": repo_diff_pb2.TectonObjectType.FEATURE_SERVICE,
    "transformation": repo_diff_pb2.TectonObjectType.TRANSFORMATION,
    "new_transformation": repo_diff_pb2.TectonObjectType.TRANSFORMATION,
}


class PlanDiff:
    def __init__(
        self,
        fco_diffs: Iterable[state_update_pb2.FcoDiff],
        warnings: Iterable[state_update_pb2.ValidationMessage],
        plan_id: Id,
    ):
        self.diffs = fco_diffs
        self.warnings = warnings
        self.plan_id = plan_id

        # preprocess warnings - group by fco id
        self.warnings_by_fco_id: Mapping[str, List[str]] = {}
        for warning in self.warnings:
            for fco_ref in warning.fco_refs:
                fco_id = IdHelper.to_string(fco_ref.fco_id)
                if fco_id not in self.warnings_by_fco_id:
                    self.warnings_by_fco_id[fco_id] = []
                self.warnings_by_fco_id[fco_id].append(warning.message)

    def print(self, debug: bool):
        """Pretty prints the plan diff in a human readable format.

        :param debug: If true, include Plan ID in output.
        """
        if not self.has_diffs():
            printer.safe_print(
                Style.BRIGHT + "🎉 The remote and local state are the same, nothing to do! ", Style.RESET_ALL
            )
            return

        printer.safe_print(Style.BRIGHT, "↓↓↓↓↓↓↓↓↓↓↓↓ Plan Start ↓↓↓↓↓↓↓↓↓↓", Style.RESET_ALL)
        printer.safe_print(Fore.RESET)
        for item in self.diffs:
            fco_id, _, _, _ = self._parse_plan_item(item)
            warnings = self._warnings_for_fco(fco_id) if fco_id else []
            self._pprint_plan_item_with_warnings(item, warnings)
            printer.safe_print(Fore.RESET)

        printer.safe_print(Style.BRIGHT, "↑↑↑↑↑↑↑↑↑↑↑↑ Plan End ↑↑↑↑↑↑↑↑↑↑↑↑", Style.RESET_ALL)
        if not debug:
            printer.safe_print(
                Style.NORMAL, "Generated plan ID is %s" % IdHelper.to_string(self.plan_id), Style.RESET_ALL
            )

        num_fcos_with_warnings = len(self.warnings_by_fco_id)
        if num_fcos_with_warnings > 0:
            printer.safe_print(
                Style.BRIGHT, f"⚠️  Plan contains warnings. Please review these warnings.", Style.RESET_ALL
            )

    def _warnings_for_fco(self, fco_id: str) -> List[str]:
        return self.warnings_by_fco_id[fco_id] if fco_id in self.warnings_by_fco_id else []

    def to_diff_summary_proto(self, debug: bool) -> repo_diff_pb2.TectonRepoDiffSummary:
        """Converts the plan diff to a summary proto.

        :param debug: If true, include Plan ID in proto.
        """
        repo_diff = repo_diff_pb2.TectonRepoDiffSummary()
        for fco_diff in self.diffs:
            repo_diff.object_diffs.append(self._convert_to_tecton_object_diff(fco_diff))
        if not debug:
            repo_diff.plan_id = IdHelper.to_string(self.plan_id)

        return repo_diff

    def confirm_plan_or_exit(self, interactive: bool, workspace: str):
        """Prompt the user to confirm the update.

        :param interactive: If true, display a yes/no prompt.
        """
        materialization_enabled = is_live_workspace(workspace)
        if materialization_enabled:
            if self._has_feature_service_update():
                printer.safe_print(
                    "Note: Updates to Feature Services may take up to 60 seconds to be propagated to the real-time feature-serving endpoint.",
                    file=sys.stderr,
                )

            if self._has_feature_view_update():
                printer.safe_print(
                    f'{Fore.YELLOW}Note: the "{workspace}" workspace is a "Live" workspace. Applying may result in new materialization jobs.{Fore.RESET}'
                )

        if interactive:
            confirm_or_exit(f'Are you sure you want to apply this plan to: "{workspace}"?')

    def has_diffs(self) -> bool:
        """Returns False if there are no changes in the diff."""
        if self.diffs:
            return True
        return False

    @property
    def num_fcos_changed(self) -> int:
        """Returns the number of FCOs being created, updated, or deleted."""
        count = 0
        for item in self.diffs:
            if item.type != state_update_pb2.FcoTransitionType.UNCHANGED:
                count += 1
        return count

    @property
    def id(self) -> Id:
        return self.plan_id

    def _has_feature_service_update(self) -> bool:
        for item in self.diffs:
            _, plan_type_item, fco_type, _ = self._parse_plan_item(item)
            if plan_type_item != "DELETE" and fco_type == "feature_service":
                return True
        return False

    def _has_feature_view_update(self) -> bool:
        for item in self.diffs:
            _, plan_type_item, fco_type, _ = self._parse_plan_item(item)
            # fco_type can be "stream_feature_view", "on_demand_feature_view", etc.
            if plan_type_item != "DELETE" and "feature_view" in fco_type:
                return True
        return False

    def _convert_to_tecton_object_diff(self, fco_diff: state_update_pb2.FcoDiff) -> repo_diff_pb2.TectonObjectDiff:
        object_diff = repo_diff_pb2.TectonObjectDiff()
        object_diff.transition_type = fco_diff.type

        fco_id, _, _, unwrapped_args = self._parse_plan_item(fco_diff)
        object_diff.object_metadata.name = unwrapped_args.info.name
        if unwrapped_args.info.owner:
            object_diff.object_metadata.owner = unwrapped_args.info.owner
        if unwrapped_args.info.description:
            object_diff.object_metadata.description = unwrapped_args.info.description

        object_diff.object_metadata.object_type = self._get_tecton_object_type(fco_diff)

        warnings = self._warnings_for_fco(fco_id) if fco_id else []
        for warning in warnings:
            object_diff.warnings.append(warning)

        return object_diff

    def _pprint_fields(self, args, colwidth):
        def pprint_metadata(indent, args):
            printer.safe_print(indent * " " + f'{"name:".ljust(colwidth)} {args.info.name}')
            if args.info.owner:
                printer.safe_print(indent * " " + f'{"owner:".ljust(colwidth)} {args.info.owner}')
            if args.info.description:
                printer.safe_print(indent * " " + f'{"description:".ljust(colwidth)} {args.info.description}')

        pprint_metadata(4, args)

    def _pprint_plan_item_with_warnings(self, item: state_update_pb2.FcoDiff, warnings: List[str]):
        fco_id, plan_item_type, fco_type, unwrapped_args = self._parse_plan_item(item)
        indent = 2
        if plan_item_type == "CREATE":
            color = Fore.GREEN
            msg = f"+ Create {human_fco_type(fco_type)}"
        elif plan_item_type == "DELETE":
            color = Fore.RED
            msg = f"- Delete {human_fco_type(fco_type)}"
        elif plan_item_type == "UPGRADE":
            color = Fore.CYAN
            msg = f"~ Upgrade {human_fco_type(fco_type)} to the latest Tecton version"
        elif plan_item_type == "UPDATE":
            color = Fore.YELLOW
            msg = f"~ Update {human_fco_type(fco_type)}"
        elif plan_item_type == "RECREATE":
            color = Fore.RED
            msg = f"~ Recreate (destructive) {human_fco_type(fco_type)}"
        else:
            assert False, f"Unknown item type {plan_item_type}"

        printer.safe_print((indent * " ") + color + msg + Fore.RESET)

        colwidth = 16
        self._pprint_fields(unwrapped_args, colwidth)

        diff_indent = 4
        for diff_item in item.diff:
            if len(diff_item.val_existing) + len(diff_item.val_declared) < 80:
                msg = diff_item.property_name + ": " + diff_item.val_existing + " -> " + diff_item.val_declared
            else:
                nl = (
                    ""
                    if len(diff_item.val_existing) + len(diff_item.val_declared) < 80
                    else "\n" + ((diff_indent + 2) * " ")
                )
                msg = (
                    diff_item.property_name
                    + ": "
                    + nl
                    + diff_item.val_existing
                    + nl
                    + " \u2B07\u2B07\u2B07\u2B07\u2B07 "
                    + nl
                    + diff_item.val_declared
                )
            printer.safe_print((diff_indent * " ") + color + msg + Fore.RESET)

        # For DELETE transitions, we do not print warnings
        if plan_item_type != "DELETE" and len(warnings) > 0:
            for warning in warnings:
                printer.safe_print(
                    4 * " "
                    + Fore.RED
                    + Style.BRIGHT
                    + f'{"warning:".ljust(colwidth)} {warning}'
                    + Fore.RESET
                    + Style.RESET_ALL
                )

    def _get_tecton_object_type(self, fco_diff: state_update_pb2.FcoDiff) -> repo_diff_pb2.TectonObjectType:
        if fco_diff.type == state_update_pb2.FcoTransitionType.CREATE:
            fco_type = fco_diff.declared_args.WhichOneof("args")
        else:
            fco_type = fco_diff.existing_args.WhichOneof("args")

        if fco_type not in _FCO_TYPE_TO_OBJECT_TYPE:
            raise errors.INTERNAL_ERROR(f"Error computing Tecton object type from FCO type: {fco_type}.")

        return _FCO_TYPE_TO_OBJECT_TYPE[fco_type]

    def _parse_plan_item(self, item: state_update_pb2.FcoDiff) -> Tuple[Optional[str], str, str, UnwrappedFcoArgs]:
        """
        :return: Parsed FcoDiff object:
                    - Client side FCO ID (Optional b/c FCO's with DELETE transition don't have client side id)
                    - Transition type
                    - FCO type
                    - Unwrapped FCO Args
        """
        plan_item_type = state_update_pb2._FCOTRANSITIONTYPE.values_by_number[item.type].name

        if item.HasField("declared_args"):  # DELETE transitions have no 'declared_args'
            fco_id, _, _ = self._unwrap_args(item.declared_args)
        else:
            fco_id = None

        if plan_item_type == "CREATE":
            fco_args = item.declared_args
        elif plan_item_type == "DELETE":
            fco_args = item.existing_args
        elif plan_item_type == "UPGRADE":
            fco_args = item.existing_args
        elif plan_item_type == "UPDATE":
            fco_args = item.existing_args
        elif plan_item_type == "RECREATE":
            fco_args = item.existing_args
        else:
            assert False, f"Unknown item type {plan_item_type}"

        _, fco_type, unwrapped_args = self._unwrap_args(fco_args)

        return fco_id, plan_item_type, fco_type, unwrapped_args

    def _unwrap_args(self, fco_args) -> Tuple[str, str, UnwrappedFcoArgs]:
        """
        :return: FCO arg properties:
                    - FCO ID
                    - FCO type
                    - UnwrappedFcoArgs object
        """
        fco_type = fco_args.WhichOneof("args")
        if fco_type == "virtual_data_source":
            unwrapped_args = fco_args.virtual_data_source
            if unwrapped_args.type == DataSourceType.BATCH:
                fco_type = "batch_data_source"
            elif unwrapped_args.type == DataSourceType.STREAM_WITH_BATCH:
                fco_type = "stream_data_source"
            fco_id = unwrapped_args.virtual_data_source_id
        elif fco_type == "entity":
            unwrapped_args = fco_args.entity
            fco_id = unwrapped_args.entity_id
        elif fco_type == "transformation":
            unwrapped_args = fco_args.transformation
            fco_id = unwrapped_args.transformation_id
        elif fco_type == "new_transformation":
            unwrapped_args = fco_args.new_transformation
            fco_id = unwrapped_args.transformation_id
        elif fco_type == "feature_view":
            unwrapped_args = fco_args.feature_view
            if unwrapped_args.feature_view_type == FeatureViewType.FEATURE_VIEW_TYPE_TEMPORAL:
                if unwrapped_args.temporal_args.data_source_type == DataSourceType.BATCH:
                    fco_type = "batch_feature_view"
                elif unwrapped_args.temporal_args.data_source_type == DataSourceType.STREAM_WITH_BATCH:
                    fco_type = "stream_feature_view"
            elif unwrapped_args.feature_view_type == FeatureViewType.FEATURE_VIEW_TYPE_TEMPORAL_AGGREGATE:
                if unwrapped_args.temporal_aggregate_args.data_source_type == DataSourceType.BATCH:
                    fco_type = "batch_window_aggregate_feature_view"
                elif unwrapped_args.temporal_aggregate_args.data_source_type == DataSourceType.STREAM_WITH_BATCH:
                    fco_type = "stream_window_aggregate_feature_view"
            elif unwrapped_args.feature_view_type == FeatureViewType.FEATURE_VIEW_TYPE_ON_DEMAND:
                fco_type = "on_demand_feature_view"
            fco_id = unwrapped_args.feature_view_id

        elif fco_type == "feature_service":
            unwrapped_args = fco_args.feature_service
            fco_id = unwrapped_args.feature_service_id
        else:
            assert False, f"Unknown object type: '{fco_type}'"

        # Currently, FeatureTable args are stored using FeatureViewArgs proto.
        # We differentiate between FeatureTable and FeatureView python classes here.
        if fco_type == "feature_view" and unwrapped_args.HasField("feature_table_args"):
            fco_type = "feature_table"

        return IdHelper.to_string(fco_id), fco_type, unwrapped_args


def get_declared_fco_args(objects) -> Tuple[List[FcoArgs], FeatureRepoSourceInfo]:
    all_args = []
    repo_source_info = FeatureRepoSourceInfo()

    for fco_obj in objects:
        fco_args = FcoArgs()
        args = fco_obj._args
        source_info = fco_obj._source_info
        if isinstance(fco_obj, BaseDataSource):
            source_info.fco_id.CopyFrom(args.virtual_data_source_id)
            fco_args.virtual_data_source.CopyFrom(args)
        elif isinstance(fco_obj, Entity):
            source_info.fco_id.CopyFrom(args.entity_id)
            fco_args.entity.CopyFrom(args)
        elif isinstance(fco_obj, Transformation):
            if isinstance(fco_obj, BuiltinTransformation) and fco_obj.call_count == 0:
                continue
            source_info.fco_id.CopyFrom(args.transformation_id)
            fco_args.new_transformation.CopyFrom(args)
        elif isinstance(fco_obj, FeatureDefinition):
            source_info.fco_id.CopyFrom(args.feature_view_id)
            fco_args.feature_view.CopyFrom(args)
        elif isinstance(fco_obj, FeatureService):
            source_info.fco_id.CopyFrom(args.feature_service_id)
            fco_args.feature_service.CopyFrom(args)
        else:
            assert False, f"Warning: unknown object{fco_obj}"

        all_args.append(fco_args)

        source_info.CopyFrom(source_info)
        repo_source_info.source_info.append(source_info)

    return all_args, repo_source_info


def dump_local_state(objects):
    with printer.safe_yaspin(Spinners.earth, text="Collecting local feature declarations") as sp:
        fco_args, repo_source_info = get_declared_fco_args(objects)
        sp.ok(printer.safe_string("✅"))

    request_plan = NewStateUpdateRequest()
    request_plan.request.fco_args.extend(fco_args)
    request_plan.request.repo_source_info.CopyFrom(repo_source_info)
    printer.safe_print(MessageToJson(request_plan, including_default_value_fields=True))


# upload tar.gz of python files to url via PUT request
def upload_files(repo_files: List[Path], repo_root, url: str):
    tar_bytes = io.BytesIO()
    with tarfile.open(fileobj=tar_bytes, mode="w|gz") as targz:
        for f in repo_files:
            targz.add(f, arcname=os.path.relpath(f, repo_root))
    for _ in range(3):
        try:
            r = requests.put(url, data=tar_bytes.getbuffer())
            if r.status_code != 200:
                # We will get 403 (forbidden) when the signed url expires.
                if r.status_code == 403:
                    printer.safe_print(
                        f"\nUploading feature repo failed due to expired session. Please retry the command."
                    )
                else:
                    printer.safe_print(f"\nUploading feature repo failed with reason: {r.reason}")
                sys.exit(1)
            return
        except requests.RequestException as e:
            last_error = e
    else:
        raise SystemExit(last_error)


def update_tecton_state(
    objects,
    repo_files: List[Path],
    repo_root: str,
    apply,
    debug,
    interactive,
    upgrade_all: bool,
    json_out_path: Optional[Path] = None,
    timeout_seconds=90 * 60,
    plan_id: Optional[str] = None,
) -> TectonStateUpdateMetrics:
    # In debug mode we compute the plan synchronously, do not save it in the database, and do not allow to apply it.
    # Primary goal is allowing local development/debugging plans against remote clusters in read-only mode.
    assert not (debug and apply), "Cannot apply in debug mode"

    workspace = get_current_workspace()

    if apply and plan_id:
        # Applying an existing plan, so skip preparing args.
        state_id = IdHelper.from_string(plan_id)
        request_query = QueryStateUpdateRequest()
        request_query.state_id.CopyFrom(state_id)
        request_query.workspace = workspace

        try:
            response_query = metadata_service.instance().QueryStateUpdate(request_query)
        except (
            tecton.tecton_errors.TectonInternalError,
            tecton._internals.tecton_errors.TectonAPIValidationError,
        ) as e:
            printer.safe_print(e)
            return TectonStateUpdateMetrics.from_error_message(str(e))

        if response_query.error:
            printer.safe_print(response_query.error)
            return TectonStateUpdateMetrics.from_error_message(response_query.error)
        if response_query.validation_result.errors:
            # Cannot pretty-print validation result using format_server_errors(), because collected local objects
            # might have changed since this plan was generated, so can't accurately match with this plan's FCOs.
            message = f"Cannot apply plan because it had errors."
            printer.safe_print(message)
            return TectonStateUpdateMetrics.from_error_message(message)
    else:
        with printer.safe_yaspin(Spinners.earth, text="Collecting local feature declarations") as sp:
            fco_args, repo_source_info = get_declared_fco_args(objects)
            sp.ok(printer.safe_string("✅"))

        request_plan = NewStateUpdateRequest()
        request_plan.request.workspace = workspace
        request_plan.request.upgrade_all = upgrade_all
        request_plan.request.sdk_version = tecton.version.get_version() or ""

        request_plan.request.fco_args.extend(fco_args)
        request_plan.request.repo_source_info.CopyFrom(repo_source_info)

        if debug:
            request_plan.blocking_dry_run_mode = True

        server_side_msg_prefix = "Performing server-side feature validation: "
        with printer.safe_yaspin(Spinners.earth, text=server_side_msg_prefix) as sp:
            try:
                response_submit = metadata_service.instance().NewStateUpdate(request_plan)
                if debug:
                    response_query = response_submit.blocking_dry_run_mode_response
                else:
                    upload_files(repo_files, repo_root, response_submit.signed_url_for_repo_upload)

                    seconds_slept = 0
                    request_query = QueryStateUpdateRequest()
                    request_query.workspace = workspace
                    request_query.state_id.CopyFrom(response_submit.state_id)
                    while True:
                        response_query = metadata_service.instance().QueryStateUpdate(request_query)
                        if response_query.latest_status_message:
                            sp.text = server_side_msg_prefix + response_query.latest_status_message
                        if response_query.ready:
                            break
                        seconds_to_sleep = 5
                        time.sleep(seconds_to_sleep)
                        seconds_slept += seconds_to_sleep
                        if seconds_slept > timeout_seconds:
                            sp.fail(printer.safe_string("⛔"))
                            printer.safe_print("Validation timed-out")
                            return TectonStateUpdateMetrics.from_error_message("Validation timed-out")

                if response_query.error:
                    sp.fail(printer.safe_string("⛔"))
                    printer.safe_print(response_query.error)
                    return TectonStateUpdateMetrics.from_error_message(response_query.error)
                if response_query.validation_result.errors:
                    sp.fail(printer.safe_string("⛔"))
                    format_server_errors(response_query.validation_result.errors, objects, repo_root)
                    return TectonStateUpdateMetrics.from_error_message(str(response_query.validation_result.errors))
                sp.ok(printer.safe_string("✅"))
            except (
                tecton.tecton_errors.TectonInternalError,
                tecton._internals.tecton_errors.TectonAPIValidationError,
            ) as e:
                sp.fail(printer.safe_string("⛔"))
                printer.safe_print(e)
                return TectonStateUpdateMetrics.from_error_message(str(e))

        state_id = response_submit.state_id

    plan_diff = PlanDiff(response_query.diff_items, response_query.validation_result.warnings, state_id)
    plan_diff.print(debug)

    if apply and plan_diff.has_diffs():
        # Use the workspace from the update request because the current workspace may have changed.
        plan_diff.confirm_plan_or_exit(interactive, workspace)

        request_apply = ApplyStateUpdateRequest()
        request_apply.state_id.CopyFrom(state_id)
        metadata_service.instance().ApplyStateUpdate(request_apply)

        printer.safe_print("🎉 all done!")

    if json_out_path:
        repo_diff_summary = plan_diff.to_diff_summary_proto(debug)
        json_out_path.parent.mkdir(parents=True, exist_ok=True)
        json_out_path.write_text(MessageToJson(repo_diff_summary))

    return TectonStateUpdateMetrics(num_total_fcos=len(objects), num_fcos_changed=plan_diff.num_fcos_changed)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/data/new_transformation.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from tecton_proto.args import user_defined_function_pb2 as tecton__proto_dot_args_dot_user__defined__function__pb2
from tecton_proto.data import fco_metadata_pb2 as tecton__proto_dot_data_dot_fco__metadata__pb2
from tecton_proto.args import new_transformation_pb2 as tecton__proto_dot_args_dot_new__transformation__pb2
from tecton_proto.common import id_pb2 as tecton__proto_dot_common_dot_id__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='tecton_proto/data/new_transformation.proto',
  package='tecton_proto.data',
  syntax='proto2',
  serialized_options=b'\n\017com.tecton.dataP\001',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n*tecton_proto/data/new_transformation.proto\x12\x11tecton_proto.data\x1a-tecton_proto/args/user_defined_function.proto\x1a$tecton_proto/data/fco_metadata.proto\x1a*tecton_proto/args/new_transformation.proto\x1a\x1ctecton_proto/common/id.proto\"\xc1\x02\n\x11NewTransformation\x12\x44\n\x11transformation_id\x18\x01 \x01(\x0b\x32\x17.tecton_proto.common.IdR\x10transformationId\x12\x41\n\x0c\x66\x63o_metadata\x18\x02 \x01(\x0b\x32\x1e.tecton_proto.data.FcoMetadataR\x0b\x66\x63oMetadata\x12V\n\x13transformation_mode\x18\x03 \x01(\x0e\x32%.tecton_proto.args.TransformationModeR\x12transformationMode\x12K\n\ruser_function\x18\x04 \x01(\x0b\x32&.tecton_proto.args.UserDefinedFunctionR\x0cuserFunctionB\x13\n\x0f\x63om.tecton.dataP\x01'
  ,
  dependencies=[tecton__proto_dot_args_dot_user__defined__function__pb2.DESCRIPTOR,tecton__proto_dot_data_dot_fco__metadata__pb2.DESCRIPTOR,tecton__proto_dot_args_dot_new__transformation__pb2.DESCRIPTOR,tecton__proto_dot_common_dot_id__pb2.DESCRIPTOR,])




_NEWTRANSFORMATION = _descriptor.Descriptor(
  name='NewTransformation',
  full_name='tecton_proto.data.NewTransformation',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='transformation_id', full_name='tecton_proto.data.NewTransformation.transformation_id', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='transformationId', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='fco_metadata', full_name='tecton_proto.data.NewTransformation.fco_metadata', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='fcoMetadata', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='transformation_mode', full_name='tecton_proto.data.NewTransformation.transformation_mode', index=2,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='transformationMode', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='user_function', full_name='tecton_proto.data.NewTransformation.user_function', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='userFunction', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=225,
  serialized_end=546,
)

_NEWTRANSFORMATION.fields_by_name['transformation_id'].message_type = tecton__proto_dot_common_dot_id__pb2._ID
_NEWTRANSFORMATION.fields_by_name['fco_metadata'].message_type = tecton__proto_dot_data_dot_fco__metadata__pb2._FCOMETADATA
_NEWTRANSFORMATION.fields_by_name['transformation_mode'].enum_type = tecton__proto_dot_args_dot_new__transformation__pb2._TRANSFORMATIONMODE
_NEWTRANSFORMATION.fields_by_name['user_function'].message_type = tecton__proto_dot_args_dot_user__defined__function__pb2._USERDEFINEDFUNCTION
DESCRIPTOR.message_types_by_name['NewTransformation'] = _NEWTRANSFORMATION
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

NewTransformation = _reflection.GeneratedProtocolMessageType('NewTransformation', (_message.Message,), {
  'DESCRIPTOR' : _NEWTRANSFORMATION,
  '__module__' : 'tecton_proto.data.new_transformation_pb2'
  # @@protoc_insertion_point(class_scope:tecton_proto.data.NewTransformation)
  })
_sym_db.RegisterMessage(NewTransformation)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



import functools
import itertools
import os
from abc import ABC
from abc import abstractmethod
from dataclasses import dataclass
from datetime import timedelta
from typing import List
from typing import Optional
from typing import Union

import pendulum
from py4j.protocol import Py4JJavaError
from pyspark.sql import Column
from pyspark.sql import DataFrame
from pyspark.sql import functions
from pyspark.sql import SparkSession
from pyspark.sql.types import IntegerType
from pyspark.sql.types import LongType
from pyspark.sql.types import StructType
from pyspark.sql.types import TimestampType

from tecton_proto.args import feature_view_pb2 as args_proto
from tecton_spark.feature_definition_wrapper import FeatureDefinitionWrapper as FeatureDefinition
from tecton_spark.logger import get_logger
from tecton_spark.partial_aggregations import TEMPORAL_ANCHOR_COLUMN_NAME
from tecton_spark.spark_time_utils import convert_epoch_to_datetime
from tecton_spark.spark_time_utils import convert_timestamp_to_epoch
from tecton_spark.time_utils import convert_timedelta_for_version

TIME_PARTITION = "time_partition"
ANCHOR_TIME = "_anchor_time"

DBRICKS_MULTI_CLUSTER_WRITES_ENABLED = "spark.databricks.delta.multiClusterWrites.enabled"
DBRICKS_RUNTIME_VERSION = "DATABRICKS_RUNTIME_VERSION"

DELTA_LOGSTORE_CLASS = "spark.delta.logStore.class"
DYNAMODB_LOGSTORE_CLASS = "io.delta.storage.DynamoDBLogStore"

logger = get_logger("offline_store")


@dataclass
class OfflineStoreParams:
    s3_path: str

    always_store_anchor_column: bool
    """Whether the anchor column should be stored in the Offline Feature Store regardless of whether it is
    required by the storage layer.

    If this is false the anchor column will be dropped from the stored data if it's not needed by the
    OfflineStoreWriter implementation.
    """

    time_column: str
    """The column containing the timestamp value used for time-based partitioning"""

    join_key_columns: List[str]

    is_continuous: bool


def get_offline_store_writer(
    params: OfflineStoreParams, fv_config: args_proto.OfflineFeatureStoreConfig, version: int, spark: SparkSession
):
    """Creates a concrete implementation of OfflineStoreWriter based on fv_config."""
    case = fv_config.WhichOneof("store_type")
    if case == "delta":
        return DeltaWriter(params, fv_config.delta, spark, version)
    elif case == "parquet":
        return ParquetWriter(params, spark, version)
    # Remove default after database migration is complete.
    # raise KeyError(case)
    return ParquetWriter(params, spark, version)


def get_offline_store_reader(spark: SparkSession, fd: FeatureDefinition) -> "OfflineStoreReader":
    case = fd.offline_store_config.WhichOneof("store_type")
    if case == "delta":
        return DeltaReader(spark, fd)
    elif case == "parquet":
        return ParquetReader(spark, fd)
    # Remove default after database migration is complete.
    # raise KeyError(case)
    return ParquetReader(spark, fd)


class OfflineStoreWriter(ABC):
    """Interface for Offline Feature Store writers."""

    @abstractmethod
    def append_dataframe(self, data_frame):
        """Append the rows from data_frame to the Store table. Nothing is overwritten."""
        raise NotImplementedError

    @abstractmethod
    def upsert_dataframe(self, data_frame):
        """Upsert the rows from data_frame to the Store table.

        Rows with matching join keys and time column are overwritten. Other rows are inserted.
        """
        raise NotImplementedError

    @abstractmethod
    def delete_keys(self, data_frame):
        """Delete rows from the Store table that match the keys inside the data_frame."""
        raise NotImplementedError


class OfflineStoreReader(ABC):
    @abstractmethod
    def read(self, time_limits: pendulum.Period) -> DataFrame:
        raise NotImplementedError


class ParquetWriter(OfflineStoreWriter):
    """Parquet implementation of OfflineStoreWriter"""

    def __init__(self, params: OfflineStoreParams, spark: SparkSession, version: int):
        self._params = params
        self._spark = spark
        self._version = version

    def append_dataframe(self, data_frame):
        if self._params.is_continuous:
            align_duration = convert_timedelta_for_version(pendulum.duration(seconds=86400), self._version)
            aligned_time = _align_timestamp(functions.col(ANCHOR_TIME), functions.lit(align_duration))
            data_frame = data_frame.withColumn(TIME_PARTITION, aligned_time)
            partition_col = TIME_PARTITION
        else:
            partition_col = ANCHOR_TIME

        mode = self._spark.conf.get("spark.sql.sources.partitionOverwriteMode")
        if mode != "dynamic":
            raise ValueError(f"partitionOverWrite mode should be dynamic, Found {mode}")
        data_frame.write.partitionBy(partition_col).parquet(self._params.s3_path, mode="overwrite")

    def upsert_dataframe(self, data_frame):
        raise NotImplementedError()

    def delete_keys(self, data_frame):
        raise NotImplementedError()


class ParquetReader(OfflineStoreReader):
    def __init__(self, spark: SparkSession, fd: FeatureDefinition):
        self._spark = spark
        assert fd.materialization_enabled and fd.writes_to_offline_store
        self._path = fd.fv_materialization.materialized_data_location.path
        self._window = fd.min_scheduling_interval
        self.version = fd.get_feature_store_format_version

    def read(self, time_limits: Optional[pendulum.Period]):
        spark_df = self._spark.read.parquet(self._path)
        if time_limits:
            start_time_epoch = convert_timestamp_to_epoch(time_limits.start, self.version)
            end_time_epoch = convert_timestamp_to_epoch(time_limits.end, self.version)
            if _size_seconds(self._window) == 0:
                start = start_time_epoch
                end = end_time_epoch
            else:
                start = _align_timestamp(start_time_epoch, _size_seconds(self._window))
                end = _align_timestamp(end_time_epoch, _size_seconds(self._window))
            anchor_time_col = functions.col(ANCHOR_TIME)
            spark_df = spark_df.where((start <= anchor_time_col) & (anchor_time_col <= end))
        return spark_df


_EXCEPTION_PACKAGES = {
    "com.databricks.sql.transaction.tahoe",  # Used by Databricks
    "org.apache.spark.sql.delta",  # Used by open source
}

_EXCEPTION_CLASSES = {
    "ConcurrentAppendException",
    "ConcurrentDeleteReadException",
    "ConcurrentDeleteDeleteException",
    "ProtocolChangedException",  # This can occur when two txns create the same table concurrently
}

_RETRYABLE_DELTA_EXCEPTIONS = {
    f"{pkg}.{cls}" for pkg, cls in itertools.product(_EXCEPTION_PACKAGES, _EXCEPTION_CLASSES)
}


def _with_delta_retries(f, max_retries=5):
    """Retries the wrapped function upon Deltalake conflict errors."""

    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        final_exception = None
        for i in range(max_retries):
            try:
                f(*args, **kwargs)
                return
            except Py4JJavaError as e:
                exception_class = e.java_exception.getClass().getCanonicalName()
                if exception_class not in _RETRYABLE_DELTA_EXCEPTIONS:
                    raise e
                final_exception = e
                logger.info(
                    f"Delta transaction failed (attempt {i + 1}/5); retrying",
                    exc_info=True,  # Include information about the exception currently being handled
                )
        raise Exception(f"Exceeded maximum Delta transaction retries ({max_retries})") from final_exception

    return wrapper


def _assert_safe_delta_write_configuration(spark: SparkSession):
    """Asserts that the Spark configuration is such that it is safe to write to Delta concurrently.

    With the Open Source Delta JAR installed (as it is on EMR), writing to a Delta table concurrently with another
    Spark cluster could corrupt the table unless the Delta Logstore class is overridden.

    On Databricks everything as fine as multi-cluster writes are enabled (the default).
    """
    configs = {
        DBRICKS_RUNTIME_VERSION: os.environ.get(DBRICKS_RUNTIME_VERSION, None),
        DBRICKS_MULTI_CLUSTER_WRITES_ENABLED: spark.conf.get(DBRICKS_MULTI_CLUSTER_WRITES_ENABLED, None),
        DELTA_LOGSTORE_CLASS: spark.conf.get(DELTA_LOGSTORE_CLASS, None),
    }
    if configs[DBRICKS_RUNTIME_VERSION] and configs[DBRICKS_MULTI_CLUSTER_WRITES_ENABLED] == "true":
        return True
    if configs[DELTA_LOGSTORE_CLASS] == DYNAMODB_LOGSTORE_CLASS:
        return True
    raise AssertionError(f"Configuration is not safe for concurrent writes: {configs}")


class DeltaWriter(OfflineStoreWriter):
    """DeltaLake implementation of OfflineStoreWriter"""

    def __init__(
        self, params: OfflineStoreParams, delta_config: args_proto.DeltaConfig, spark: SparkSession, version: int
    ):
        self._params = params
        self._delta_config = delta_config
        self._version = version
        self._spark = spark
        self._window = delta_config.time_partition_size.ToTimedelta()
        if not spark.conf.get("spark.databricks.delta.commitInfo.userMetadata"):
            raise AssertionError(f"Expected spark.databricks.delta.commitInfo.userMetadata to be set for delta writes")

    def append_dataframe(self, data_frame: DataFrame):
        data_frame = self._add_partition(data_frame)
        self._ensure_table_exists(self._spark, data_frame.schema)
        self._append_dataframe(data_frame)

    def upsert_dataframe(self, data_frame):
        # See https://github.com/delta-io/delta/issues/282 for why this isn't at the top of the file
        from delta.tables import DeltaTable

        _assert_safe_delta_write_configuration(self._spark)

        data_frame = self._add_partition(data_frame)
        self._ensure_table_exists(self._spark, data_frame.schema)

        table = DeltaTable.forPath(self._spark, self._params.s3_path)

        base = table.toDF().alias("base")
        updates = data_frame.alias("updates")

        # Build a condition which matches on all join keys, the timestamp, and the time partition column. The time
        # partition column is not needed for correctness, but it allows some files to be skipped by Delta.
        all_match_keys = [self._params.time_column, TIME_PARTITION, *self._params.join_key_columns]
        key_matches = [base[k] == updates[k] for k in all_match_keys]
        match_condition = functools.reduce(lambda l, r: l & r, key_matches)

        @_with_delta_retries
        def _execute():
            table.merge(updates, match_condition).whenMatchedUpdateAll().whenNotMatchedInsertAll().execute()

        _execute()

    def delete_keys(self, data_frame):
        # See https://github.com/delta-io/delta/issues/282 for why this isn't at the top of the file
        from delta.tables import DeltaTable

        _assert_safe_delta_write_configuration(self._spark)

        deltaTable = DeltaTable.forPath(self._spark, self._params.s3_path)
        query = ""
        columns = data_frame.columns
        for column in columns:
            if query:
                query = query + " AND "
            query = query + "t." + column + " = k." + column

        @_with_delta_retries
        def _execute():
            deltaTable.alias("t").merge(data_frame.alias("k"), query).whenMatchedDelete().execute()

        @_with_delta_retries
        def _vacuum():
            deltaTable.vacuum()

        _execute()
        _vacuum()

    def delete_time_range(self, feature_start_time, feature_end_time):
        from delta.tables import DeltaTable

        start_timestamp = pendulum.instance(feature_start_time.ToDatetime()).int_timestamp
        end_timestamp = pendulum.instance(feature_end_time.ToDatetime()).int_timestamp
        start_partition = _unix_timestamp_to_partition(start_timestamp, self._window)
        end_partition = _unix_timestamp_to_partition(end_timestamp, self._window)
        start_time = feature_start_time.ToNanoseconds()
        end_time = feature_end_time.ToNanoseconds()

        assert self._params.always_store_anchor_column

        table = DeltaTable.forPath(self._spark, self._params.s3_path)

        @_with_delta_retries
        def _execute():
            table.delete(
                f'{ANCHOR_TIME} >= {start_time} and {ANCHOR_TIME} < {end_time} and {TIME_PARTITION} >= to_timestamp("{start_partition}") and {TIME_PARTITION} <= to_timestamp("{end_partition}")'
            )

        _execute()

    def _add_partition(self, data_frame: DataFrame) -> DataFrame:
        """Adds the time_partition column and drops the _anchor_time column if needed."""
        partition_size = self._delta_config.time_partition_size.ToTimedelta()
        partition = _timestamp_to_partition_column(data_frame, self._params.time_column, partition_size, self._version)
        data_frame = data_frame.withColumn(TIME_PARTITION, partition)
        if not self._params.always_store_anchor_column:
            data_frame = data_frame.drop(ANCHOR_TIME)
        return data_frame

    def _ensure_table_exists(self, spark: SparkSession, schema: StructType):
        """Ensures that the table exists with the given schema.

        Some operations (including merge) fail when the table doesn't already exist. Others (append) can have conflicts
        where they wouldn't normally when they also create a new table. This function ensures neither will happen.
        """
        df = spark.createDataFrame([], schema)  # DF with 0 rows
        self._append_dataframe(df)

    @_with_delta_retries
    def _append_dataframe(self, df: DataFrame):
        _assert_safe_delta_write_configuration(self._spark)
        df.write.partitionBy(TIME_PARTITION).format("delta").mode("append").save(self._params.s3_path)


class DeltaReader(OfflineStoreReader):
    def __init__(self, spark: SparkSession, fd: FeatureDefinition):
        self._spark = spark
        assert fd.materialization_enabled and fd.writes_to_offline_store
        self._path = fd.fv_materialization.materialized_data_location.path
        self._window = fd.offline_store_config.delta.time_partition_size.ToTimedelta()

    def read(self, time_limits: Optional[pendulum.Period]):
        spark_df = self._spark.read.format("delta").load(self._path)
        if time_limits is not None:
            start = _unix_timestamp_to_partition(time_limits.start.int_timestamp, self._window)
            end = _unix_timestamp_to_partition(time_limits.end.int_timestamp, self._window)
            partition_col = functions.col(TIME_PARTITION)
            spark_df = spark_df.where((start <= partition_col) & (partition_col <= end))
        return spark_df

    def history(self):
        from delta.tables import DeltaTable

        delta_table = DeltaTable.forPath(spark, materialization_params.HasField("s3_output_path"))
        return delta_table.history()


@dataclass
class TimestampFormats:
    spark_format: str
    python_format: str


def _size_seconds(window: Union[timedelta, pendulum.Duration]):
    if isinstance(window, pendulum.Duration):
        window = window.as_timedelta()
    if window % timedelta(seconds=1) != timedelta(0):
        raise AssertionError(f"partition_size is not a round number of seconds: {window}")
    return int(window.total_seconds())


def _timestamp_formats(partition_size: timedelta):
    if partition_size % timedelta(days=1) == timedelta(0):
        return TimestampFormats(spark_format="yyyy-MM-dd", python_format="%Y-%m-%d")
    else:
        return TimestampFormats(spark_format="yyyy-MM-dd-HH:mm:ss", python_format="%Y-%m-%d-%H:%M:%S")


def _timestamp_to_partition_column(df: DataFrame, time_col: str, partition_size: timedelta, version: int) -> Column:
    # For some insane reason from_unixtime returns a timestamp in the session timezone, so it's pretty annoying to
    # convert a unix time to a formatted UTC timestamp unless the session is set to UTC. This only runs in
    # materialization so we can just assert that that's the case.
    tz = df.sql_ctx.sparkSession.conf.get("spark.sql.session.timeZone")
    if tz not in {"UTC", "Etc/UTC", "GMT"}:
        raise AssertionError(f"spark.sql.session.timeZone must be UTC, not {tz}")
    time_column_type = df.schema[time_col].dataType
    allowed_types = {IntegerType(), TimestampType(), LongType()}
    if time_column_type not in allowed_types:
        raise AssertionError(f"timestamp column must be one of {allowed_types}, not {time_column_type}")
    time_val = functions.col(time_col).cast(LongType())
    if time_col == TEMPORAL_ANCHOR_COLUMN_NAME:
        time_val = convert_epoch_to_datetime(functions.col(time_col), version).cast(LongType())
    aligned = functions.from_unixtime(_align_timestamp(time_val, _size_seconds(partition_size)))
    partition_format = _timestamp_formats(partition_size).spark_format
    return functions.date_format(aligned.cast(TimestampType()), partition_format)


def _unix_timestamp_to_partition(unix_timestamp: int, partition_size: timedelta) -> str:
    dt = pendulum.from_timestamp(_align_timestamp(unix_timestamp, _size_seconds(partition_size)))
    partition_format = _timestamp_formats(partition_size).python_format
    return dt.strftime(partition_format)


def _align_timestamp(int_timestamp_col, window_size):
    return int_timestamp_col - (int_timestamp_col % window_size)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/data/serving_status.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from tecton_proto.common import id_pb2 as tecton__proto_dot_common_dot_id__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='tecton_proto/data/serving_status.proto',
  package='tecton_proto.data',
  syntax='proto2',
  serialized_options=b'\n\017com.tecton.dataP\001',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n&tecton_proto/data/serving_status.proto\x12\x11tecton_proto.data\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x1ctecton_proto/common/id.proto\"m\n\rServingStatus\x12\x44\n\rserving_state\x18\x01 \x01(\x0e\x32\x1f.tecton_proto.data.ServingStateR\x0cservingState\x12\x16\n\x06\x65rrors\x18\x02 \x03(\tR\x06\x65rrors\"\xcd\x01\n\x0bStatusRange\x12\x43\n\x0f\x62\x65gin_inclusive\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x0e\x62\x65ginInclusive\x12?\n\rend_exclusive\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x0c\x65ndExclusive\x12\x38\n\x06status\x18\x03 \x01(\x0b\x32 .tecton_proto.data.ServingStatusR\x06status\"\xc5\x01\n FeatureViewMaterializationRanges\x12?\n\x0f\x66\x65\x61ture_view_id\x18\x01 \x01(\x0b\x32\x17.tecton_proto.common.IdR\rfeatureViewId\x12`\n\x1c\x62\x61tch_materialization_ranges\x18\x02 \x03(\x0b\x32\x1e.tecton_proto.data.StatusRangeR\x1a\x62\x61tchMaterializationRanges\"\xc8\x04\n\x14ServingStatusSummary\x12X\n\x18offline_readiness_ranges\x18\x01 \x03(\x0b\x32\x1e.tecton_proto.data.StatusRangeR\x16offlineReadinessRanges\x12u\n\x1c\x62\x61tch_materialization_ranges\x18\x04 \x03(\x0b\x32\x33.tecton_proto.data.FeatureViewMaterializationRangesR\x1a\x62\x61tchMaterializationRanges\x12L\n\x14streaming_start_time\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x12streamingStartTime\x12Z\n\x18streaming_serving_status\x18\x02 \x01(\x0b\x32 .tecton_proto.data.ServingStatusR\x16streamingServingStatus\x12O\n\x16most_recent_ready_time\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x13mostRecentReadyTime\x12\x64\n!most_recent_batch_processing_time\x18\x06 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x1dmostRecentBatchProcessingTime\"\xc7\x01\n\x1f\x46\x65\x61tureViewServingStatusSummary\x12?\n\x0f\x66\x65\x61ture_view_id\x18\x03 \x01(\x0b\x32\x17.tecton_proto.common.IdR\rfeatureViewId\x12]\n\x16serving_status_summary\x18\x02 \x01(\x0b\x32\'.tecton_proto.data.ServingStatusSummaryR\x14servingStatusSummaryJ\x04\x08\x01\x10\x02\"\xa6\x02\n FullFeatureServiceServingSummary\x12{\n&feature_service_serving_status_summary\x18\x02 \x01(\x0b\x32\'.tecton_proto.data.ServingStatusSummaryR\"featureServiceServingStatusSummary\x12\x84\x01\n%feature_view_serving_status_summaries\x18\x03 \x03(\x0b\x32\x32.tecton_proto.data.FeatureViewServingStatusSummaryR!featureViewServingStatusSummaries*\xcd\x01\n\x0cServingState\x12\x19\n\x15SERVING_STATE_UNKNOWN\x10\x00\x12\x1a\n\x16SERVING_STATE_DISABLED\x10\x01\x12!\n\x1dSERVING_STATE_NOT_ENOUGH_DATA\x10\x02\x12\x19\n\x15SERVING_STATE_PENDING\x10\x03\x12\x14\n\x10SERVING_STATE_OK\x10\x04\x12\x17\n\x13SERVING_STATE_ERROR\x10\x05\x12\x19\n\x15SERVING_STATE_RUNNING\x10\x06\x42\x13\n\x0f\x63om.tecton.dataP\x01'
  ,
  dependencies=[google_dot_protobuf_dot_timestamp__pb2.DESCRIPTOR,tecton__proto_dot_common_dot_id__pb2.DESCRIPTOR,])

_SERVINGSTATE = _descriptor.EnumDescriptor(
  name='ServingState',
  full_name='tecton_proto.data.ServingState',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='SERVING_STATE_UNKNOWN', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SERVING_STATE_DISABLED', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SERVING_STATE_NOT_ENOUGH_DATA', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SERVING_STATE_PENDING', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SERVING_STATE_OK', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SERVING_STATE_ERROR', index=5, number=5,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SERVING_STATE_RUNNING', index=6, number=6,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=1730,
  serialized_end=1935,
)
_sym_db.RegisterEnumDescriptor(_SERVINGSTATE)

ServingState = enum_type_wrapper.EnumTypeWrapper(_SERVINGSTATE)
SERVING_STATE_UNKNOWN = 0
SERVING_STATE_DISABLED = 1
SERVING_STATE_NOT_ENOUGH_DATA = 2
SERVING_STATE_PENDING = 3
SERVING_STATE_OK = 4
SERVING_STATE_ERROR = 5
SERVING_STATE_RUNNING = 6



_SERVINGSTATUS = _descriptor.Descriptor(
  name='ServingStatus',
  full_name='tecton_proto.data.ServingStatus',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='serving_state', full_name='tecton_proto.data.ServingStatus.serving_state', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='servingState', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='errors', full_name='tecton_proto.data.ServingStatus.errors', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='errors', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=124,
  serialized_end=233,
)


_STATUSRANGE = _descriptor.Descriptor(
  name='StatusRange',
  full_name='tecton_proto.data.StatusRange',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='begin_inclusive', full_name='tecton_proto.data.StatusRange.begin_inclusive', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='beginInclusive', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='end_exclusive', full_name='tecton_proto.data.StatusRange.end_exclusive', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='endExclusive', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='status', full_name='tecton_proto.data.StatusRange.status', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='status', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=236,
  serialized_end=441,
)


_FEATUREVIEWMATERIALIZATIONRANGES = _descriptor.Descriptor(
  name='FeatureViewMaterializationRanges',
  full_name='tecton_proto.data.FeatureViewMaterializationRanges',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='feature_view_id', full_name='tecton_proto.data.FeatureViewMaterializationRanges.feature_view_id', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='featureViewId', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='batch_materialization_ranges', full_name='tecton_proto.data.FeatureViewMaterializationRanges.batch_materialization_ranges', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='batchMaterializationRanges', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=444,
  serialized_end=641,
)


_SERVINGSTATUSSUMMARY = _descriptor.Descriptor(
  name='ServingStatusSummary',
  full_name='tecton_proto.data.ServingStatusSummary',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='offline_readiness_ranges', full_name='tecton_proto.data.ServingStatusSummary.offline_readiness_ranges', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='offlineReadinessRanges', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='batch_materialization_ranges', full_name='tecton_proto.data.ServingStatusSummary.batch_materialization_ranges', index=1,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='batchMaterializationRanges', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='streaming_start_time', full_name='tecton_proto.data.ServingStatusSummary.streaming_start_time', index=2,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='streamingStartTime', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='streaming_serving_status', full_name='tecton_proto.data.ServingStatusSummary.streaming_serving_status', index=3,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='streamingServingStatus', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='most_recent_ready_time', full_name='tecton_proto.data.ServingStatusSummary.most_recent_ready_time', index=4,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='mostRecentReadyTime', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='most_recent_batch_processing_time', full_name='tecton_proto.data.ServingStatusSummary.most_recent_batch_processing_time', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='mostRecentBatchProcessingTime', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=644,
  serialized_end=1228,
)


_FEATUREVIEWSERVINGSTATUSSUMMARY = _descriptor.Descriptor(
  name='FeatureViewServingStatusSummary',
  full_name='tecton_proto.data.FeatureViewServingStatusSummary',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='feature_view_id', full_name='tecton_proto.data.FeatureViewServingStatusSummary.feature_view_id', index=0,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='featureViewId', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='serving_status_summary', full_name='tecton_proto.data.FeatureViewServingStatusSummary.serving_status_summary', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='servingStatusSummary', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1231,
  serialized_end=1430,
)


_FULLFEATURESERVICESERVINGSUMMARY = _descriptor.Descriptor(
  name='FullFeatureServiceServingSummary',
  full_name='tecton_proto.data.FullFeatureServiceServingSummary',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='feature_service_serving_status_summary', full_name='tecton_proto.data.FullFeatureServiceServingSummary.feature_service_serving_status_summary', index=0,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='featureServiceServingStatusSummary', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='feature_view_serving_status_summaries', full_name='tecton_proto.data.FullFeatureServiceServingSummary.feature_view_serving_status_summaries', index=1,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='featureViewServingStatusSummaries', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1433,
  serialized_end=1727,
)

_SERVINGSTATUS.fields_by_name['serving_state'].enum_type = _SERVINGSTATE
_STATUSRANGE.fields_by_name['begin_inclusive'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_STATUSRANGE.fields_by_name['end_exclusive'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_STATUSRANGE.fields_by_name['status'].message_type = _SERVINGSTATUS
_FEATUREVIEWMATERIALIZATIONRANGES.fields_by_name['feature_view_id'].message_type = tecton__proto_dot_common_dot_id__pb2._ID
_FEATUREVIEWMATERIALIZATIONRANGES.fields_by_name['batch_materialization_ranges'].message_type = _STATUSRANGE
_SERVINGSTATUSSUMMARY.fields_by_name['offline_readiness_ranges'].message_type = _STATUSRANGE
_SERVINGSTATUSSUMMARY.fields_by_name['batch_materialization_ranges'].message_type = _FEATUREVIEWMATERIALIZATIONRANGES
_SERVINGSTATUSSUMMARY.fields_by_name['streaming_start_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_SERVINGSTATUSSUMMARY.fields_by_name['streaming_serving_status'].message_type = _SERVINGSTATUS
_SERVINGSTATUSSUMMARY.fields_by_name['most_recent_ready_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_SERVINGSTATUSSUMMARY.fields_by_name['most_recent_batch_processing_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_FEATUREVIEWSERVINGSTATUSSUMMARY.fields_by_name['feature_view_id'].message_type = tecton__proto_dot_common_dot_id__pb2._ID
_FEATUREVIEWSERVINGSTATUSSUMMARY.fields_by_name['serving_status_summary'].message_type = _SERVINGSTATUSSUMMARY
_FULLFEATURESERVICESERVINGSUMMARY.fields_by_name['feature_service_serving_status_summary'].message_type = _SERVINGSTATUSSUMMARY
_FULLFEATURESERVICESERVINGSUMMARY.fields_by_name['feature_view_serving_status_summaries'].message_type = _FEATUREVIEWSERVINGSTATUSSUMMARY
DESCRIPTOR.message_types_by_name['ServingStatus'] = _SERVINGSTATUS
DESCRIPTOR.message_types_by_name['StatusRange'] = _STATUSRANGE
DESCRIPTOR.message_types_by_name['FeatureViewMaterializationRanges'] = _FEATUREVIEWMATERIALIZATIONRANGES
DESCRIPTOR.message_types_by_name['ServingStatusSummary'] = _SERVINGSTATUSSUMMARY
DESCRIPTOR.message_types_by_name['FeatureViewServingStatusSummary'] = _FEATUREVIEWSERVINGSTATUSSUMMARY
DESCRIPTOR.message_types_by_name['FullFeatureServiceServingSummary'] = _FULLFEATURESERVICESERVINGSUMMARY
DESCRIPTOR.enum_types_by_name['ServingState'] = _SERVINGSTATE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ServingStatus = _reflection.GeneratedProtocolMessageType('ServingStatus', (_message.Message,), {
  'DESCRIPTOR' : _SERVINGSTATUS,
  '__module__' : 'tecton_proto.data.serving_status_pb2'
  # @@protoc_insertion_point(class_scope:tecton_proto.data.ServingStatus)
  })
_sym_db.RegisterMessage(ServingStatus)

StatusRange = _reflection.GeneratedProtocolMessageType('StatusRange', (_message.Message,), {
  'DESCRIPTOR' : _STATUSRANGE,
  '__module__' : 'tecton_proto.data.serving_status_pb2'
  # @@protoc_insertion_point(class_scope:tecton_proto.data.StatusRange)
  })
_sym_db.RegisterMessage(StatusRange)

FeatureViewMaterializationRanges = _reflection.GeneratedProtocolMessageType('FeatureViewMaterializationRanges', (_message.Message,), {
  'DESCRIPTOR' : _FEATUREVIEWMATERIALIZATIONRANGES,
  '__module__' : 'tecton_proto.data.serving_status_pb2'
  # @@protoc_insertion_point(class_scope:tecton_proto.data.FeatureViewMaterializationRanges)
  })
_sym_db.RegisterMessage(FeatureViewMaterializationRanges)

ServingStatusSummary = _reflection.GeneratedProtocolMessageType('ServingStatusSummary', (_message.Message,), {
  'DESCRIPTOR' : _SERVINGSTATUSSUMMARY,
  '__module__' : 'tecton_proto.data.serving_status_pb2'
  # @@protoc_insertion_point(class_scope:tecton_proto.data.ServingStatusSummary)
  })
_sym_db.RegisterMessage(ServingStatusSummary)

FeatureViewServingStatusSummary = _reflection.GeneratedProtocolMessageType('FeatureViewServingStatusSummary', (_message.Message,), {
  'DESCRIPTOR' : _FEATUREVIEWSERVINGSTATUSSUMMARY,
  '__module__' : 'tecton_proto.data.serving_status_pb2'
  # @@protoc_insertion_point(class_scope:tecton_proto.data.FeatureViewServingStatusSummary)
  })
_sym_db.RegisterMessage(FeatureViewServingStatusSummary)

FullFeatureServiceServingSummary = _reflection.GeneratedProtocolMessageType('FullFeatureServiceServingSummary', (_message.Message,), {
  'DESCRIPTOR' : _FULLFEATURESERVICESERVINGSUMMARY,
  '__module__' : 'tecton_proto.data.serving_status_pb2'
  # @@protoc_insertion_point(class_scope:tecton_proto.data.FullFeatureServiceServingSummary)
  })
_sym_db.RegisterMessage(FullFeatureServiceServingSummary)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/args/new_transformation.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from tecton_proto.args import user_defined_function_pb2 as tecton__proto_dot_args_dot_user__defined__function__pb2
from tecton_proto.args import basic_info_pb2 as tecton__proto_dot_args_dot_basic__info__pb2
from tecton_proto.common import id_pb2 as tecton__proto_dot_common_dot_id__pb2
from tecton_proto.args import diff_options_pb2 as tecton__proto_dot_args_dot_diff__options__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='tecton_proto/args/new_transformation.proto',
  package='tecton_proto.args',
  syntax='proto2',
  serialized_options=b'\n\017com.tecton.argsP\001',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n*tecton_proto/args/new_transformation.proto\x12\x11tecton_proto.args\x1a-tecton_proto/args/user_defined_function.proto\x1a\"tecton_proto/args/basic_info.proto\x1a\x1ctecton_proto/common/id.proto\x1a$tecton_proto/args/diff_options.proto\"\xff\x02\n\x15NewTransformationArgs\x12\x44\n\x11transformation_id\x18\x01 \x01(\x0b\x32\x17.tecton_proto.common.IdR\x10transformationId\x12\x37\n\x04info\x18\x02 \x01(\x0b\x32\x1c.tecton_proto.args.BasicInfoB\x05\x92M\x02\x10\x01R\x04info\x12V\n\x13transformation_mode\x18\x03 \x01(\x0e\x32%.tecton_proto.args.TransformationModeR\x12transformationMode\x12K\n\ruser_function\x18\x04 \x01(\x0b\x32&.tecton_proto.args.UserDefinedFunctionR\x0cuserFunction\x12#\n\tdocstring\x18\x06 \x01(\tB\x05\x92M\x02\x08\x01R\tdocstring\x12\x1d\n\nis_builtin\x18\x05 \x01(\x08R\tisBuiltin*\xe0\x01\n\x12TransformationMode\x12\x1f\n\x1bTRANSFORMATION_MODE_UNKNOWN\x10\x00\x12\x1f\n\x1bTRANSFORMATION_MODE_PYSPARK\x10\x01\x12!\n\x1dTRANSFORMATION_MODE_SPARK_SQL\x10\x02\x12\x1e\n\x1aTRANSFORMATION_MODE_PANDAS\x10\x03\x12%\n!TRANSFORMATION_MODE_SNOWFLAKE_SQL\x10\x04\x12\x1e\n\x1aTRANSFORMATION_MODE_PYTHON\x10\x05\x42\x13\n\x0f\x63om.tecton.argsP\x01'
  ,
  dependencies=[tecton__proto_dot_args_dot_user__defined__function__pb2.DESCRIPTOR,tecton__proto_dot_args_dot_basic__info__pb2.DESCRIPTOR,tecton__proto_dot_common_dot_id__pb2.DESCRIPTOR,tecton__proto_dot_args_dot_diff__options__pb2.DESCRIPTOR,])

_TRANSFORMATIONMODE = _descriptor.EnumDescriptor(
  name='TransformationMode',
  full_name='tecton_proto.args.TransformationMode',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='TRANSFORMATION_MODE_UNKNOWN', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TRANSFORMATION_MODE_PYSPARK', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TRANSFORMATION_MODE_SPARK_SQL', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TRANSFORMATION_MODE_PANDAS', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TRANSFORMATION_MODE_SNOWFLAKE_SQL', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TRANSFORMATION_MODE_PYTHON', index=5, number=5,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=603,
  serialized_end=827,
)
_sym_db.RegisterEnumDescriptor(_TRANSFORMATIONMODE)

TransformationMode = enum_type_wrapper.EnumTypeWrapper(_TRANSFORMATIONMODE)
TRANSFORMATION_MODE_UNKNOWN = 0
TRANSFORMATION_MODE_PYSPARK = 1
TRANSFORMATION_MODE_SPARK_SQL = 2
TRANSFORMATION_MODE_PANDAS = 3
TRANSFORMATION_MODE_SNOWFLAKE_SQL = 4
TRANSFORMATION_MODE_PYTHON = 5



_NEWTRANSFORMATIONARGS = _descriptor.Descriptor(
  name='NewTransformationArgs',
  full_name='tecton_proto.args.NewTransformationArgs',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='transformation_id', full_name='tecton_proto.args.NewTransformationArgs.transformation_id', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='transformationId', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='info', full_name='tecton_proto.args.NewTransformationArgs.info', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\222M\002\020\001', json_name='info', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='transformation_mode', full_name='tecton_proto.args.NewTransformationArgs.transformation_mode', index=2,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='transformationMode', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='user_function', full_name='tecton_proto.args.NewTransformationArgs.user_function', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='userFunction', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='docstring', full_name='tecton_proto.args.NewTransformationArgs.docstring', index=4,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\222M\002\010\001', json_name='docstring', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='is_builtin', full_name='tecton_proto.args.NewTransformationArgs.is_builtin', index=5,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='isBuiltin', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=217,
  serialized_end=600,
)

_NEWTRANSFORMATIONARGS.fields_by_name['transformation_id'].message_type = tecton__proto_dot_common_dot_id__pb2._ID
_NEWTRANSFORMATIONARGS.fields_by_name['info'].message_type = tecton__proto_dot_args_dot_basic__info__pb2._BASICINFO
_NEWTRANSFORMATIONARGS.fields_by_name['transformation_mode'].enum_type = _TRANSFORMATIONMODE
_NEWTRANSFORMATIONARGS.fields_by_name['user_function'].message_type = tecton__proto_dot_args_dot_user__defined__function__pb2._USERDEFINEDFUNCTION
DESCRIPTOR.message_types_by_name['NewTransformationArgs'] = _NEWTRANSFORMATIONARGS
DESCRIPTOR.enum_types_by_name['TransformationMode'] = _TRANSFORMATIONMODE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

NewTransformationArgs = _reflection.GeneratedProtocolMessageType('NewTransformationArgs', (_message.Message,), {
  'DESCRIPTOR' : _NEWTRANSFORMATIONARGS,
  '__module__' : 'tecton_proto.args.new_transformation_pb2'
  # @@protoc_insertion_point(class_scope:tecton_proto.args.NewTransformationArgs)
  })
_sym_db.RegisterMessage(NewTransformationArgs)


DESCRIPTOR._options = None
_NEWTRANSFORMATIONARGS.fields_by_name['info']._options = None
_NEWTRANSFORMATIONARGS.fields_by_name['docstring']._options = None
# @@protoc_insertion_point(module_scope)



class Schema:
    def __init__(self, proto):
        self._proto = proto

    def to_proto(self):
        return self._proto

    def tecton_type(self, column_name):
        c = self._column(column_name)
        return c.feature_server_type if c.HasField("feature_server_type") else None

    def spark_type(self, column_name):
        c = self._column(column_name)
        if not c.HasField("raw_spark_type"):
            raise ValueError(f"Column {column_name} is missing raw_spark_type")
        return c.raw_spark_type

    def _column(self, column_name):
        cs = [c for c in self._proto.columns if c.name == column_name]
        if not cs:
            raise ValueError(f"Unknown column: {column_name}. Schema is: {self._proto}")
        return cs[0]

    def column_names(self):
        return [c.name for c in self._proto.columns]

    def column_name_raw_spark_types(self):
        return [(c.name, c.raw_spark_type) for c in self._proto.columns]



from abc import ABC
from abc import abstractmethod

from tecton_proto.args import virtual_data_source_pb2


class BaseBatchDSConfig(ABC):
    @abstractmethod
    def _merge_batch_args(self, data_source_args: virtual_data_source_pb2.VirtualDataSourceArgs):
        pass


class BaseStreamDSConfig(ABC):
    @abstractmethod
    def _merge_stream_args(self, data_source_args: virtual_data_source_pb2.VirtualDataSourceArgs):
        pass



BUILD_STATUS = {'BUILD_ID': 'uncommitted-buildkite-agent-1646850889', 'BUILD_TIMESTAMP': '1646850888', 'GIT_COMMIT': '759be48e5ee0e01be2d1b2368e03cb04c1694d25', 'BUILD_EMBED_LABEL': '', 'BUILD_HOST': 'buildkite-agent-default-13', 'BUILD_USER': 'buildkite-agent'}
